/*
* BearLibTerminal
* Copyright (C) 2013 Cfyz
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
* FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
* COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
* IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
* CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#define BEARLIBTERMINAL_LIBRARY

#include <stdarg.h>
#include <cstring>
#include <cstdio>
#include <memory>
#include "BearLibTerminal.h"
#include "Terminal.hpp"
#include "Encoding.hpp"
#include "Utility.hpp"
#include "Palette.hpp"
#include "Log.hpp"

namespace BearLibTerminal
{
	std::unique_ptr<Terminal> instance;
}

using namespace BearLibTerminal;

int terminal_open(int codepage)
{
	if ( instance ) terminal_close();
	if ( codepage == 0 ) codepage = CODEPAGE_DEFAULT;

	try
	{
		instance.reset(new Terminal(codepage));
		return 0;
	}
	catch ( std::exception& )
	{
		// ...
	}

	return -1;
}

void terminal_close()
{
	instance.reset();
}

int terminal_options(const char* options, ...)
{
	if ( !instance ) return -1;
	
	// NOTE: using static storage makes this method non-reentrant and thus not thread-safe
	static const size_t buffer_size = 1024;
	static char buffer[buffer_size] = { 0 };

	if ( options == nullptr )
	{
		// TODO: LOG
		return -1;
	}

	va_list argptr;
	va_start(argptr, options);
	int rc = std::vsnprintf(buffer, buffer_size-1, options, argptr);
	va_end(argptr);

	if ( rc < 0 )
	{
		// TODO: LOG
		return -1;
	}

	return instance->SetOptions(buffer);
}

int terminal_options_noformat(const char* options)
{
	if ( !instance ) return -1;
	return instance->SetOptions(options);
}

void terminal_refresh()
{
	if ( instance ) instance->Refresh();
}

void terminal_clear()
{
	if ( instance ) instance->Clear();
}

void terminal_clear_area(int x, int y, int w, int h)
{
	if ( instance ) instance->Clear(x, y, w, h);
}

void terminal_color(color_t color)
{
	if ( instance ) instance->SetColor(color);
}

void terminal_bkcolor(color_t color)
{
	if ( instance ) instance->SetBkColor(color);
}

void terminal_blending(int mode)
{
	if ( instance ) instance->SetBlending(mode);
}

void terminal_put(int x, int y, int code)
{
	if ( instance ) instance->Put(x, y, code);
}

size_t terminal_printf(int x, int y, const char* format, ...)
{
	if ( !instance ) return -1;

	// NOTE: using static storage makes this method non-reentrant and thus not thread-safe
	// FIXME: buffer reallocation depending on the string being formatted
	static const size_t buffer_size = 1024;
	static char buffer[buffer_size] = { 0 };

	if ( format == nullptr )
	{
		// TODO: LOG
		return -1;
	}

	va_list argptr;
	va_start(argptr, format);
	int rc = std::vsnprintf(buffer, buffer_size-1, format, argptr);
	va_end(argptr);

	if ( rc < 0 )
	{
		// TODO: LOG
		return -1;
	}

	return instance->Print(x, y, buffer);
}

size_t terminal_printf_noformat(int x, int y, const char* s)
{
	if ( !instance ) return -1;

	if ( s == nullptr )
	{
		// TODO: LOG
		return -1;
	}

	return instance->Print(x, y, s);
}

int terminal_has_input()
{
	if ( !instance ) return 1; // So that next terminal_get call will result in immediate VK_CLOSE
	return instance->HasInput();
}

int terminal_state(int code)
{
	if ( !instance ) return 0;
	return instance->GetState(code);
}

int terminal_get()
{
	if ( !instance ) return VK_CLOSE;
	return instance->Get();
}

int terminal_getstr(int x, int y, char* buffer, int max)
{
	if ( !instance ) return -1;
	return instance->GetStr(x, y, buffer, max);
}

color_t color_from_name(const char* name)
{
	if ( name == nullptr ) return COLOR_BLACK;

	// Convert latin (and so ASCII) name to wide string
	size_t length = strlen(name);
	std::wstring name_u16(length, L'.');
	for ( size_t i = 0; i < length; i++ ) name_u16[i] = (wchar_t)name[i];

	// Retrieve
	return (color_t)Palette::Instance[name_u16];
}

int terminal_use_image(int code, const char* filename)
{
	if (code < 0 || code > 0xFFFF || !instance) return -1;
	return instance->UseImage(code, filename);
}
