/*
* BearLibTerminal
* Copyright (C) 2013 Cfyz
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
* FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
* COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
* IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
* CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include <stdexcept>
#include <exception>
#include <fstream>
#include <cstdint>
#include "BitmapFontProvider.hpp"
#include "Utility.hpp"
#include "Geometry.hpp"
#include "LoadBitmap.hpp"
#include "Resources.hpp"
#include "Log.hpp"

namespace BearLibTerminal
{
	BitmapFontProvider::BitmapFontProvider(const FontProvider::Options& options): FontProvider(options)
	{
		if ( options.size.width == 0 || options.size.height == 0 )
		{
			throw std::runtime_error("invalid font size");
		}

		m_cell_size = m_glyph_size = options.size;

		try
		{
			m_encoding = Encoding(options.codepage);
		}
		catch ( std::invalid_argument& )
		{
			throw std::runtime_error("invalid font codepage name");
		}

		m_cache = LoadBitmap(*Resources::OpenStream(options.name));
	}

	BitmapFontProvider::~BitmapFontProvider()
	{ }

	void BitmapFontProvider::Dispose() // FIXME: unnecessary
	{ }

	Bitmap BitmapFontProvider::GetGlyphBitmap(wchar_t code)
	{ 
		return FontProvider::GetGlyphBitmap(code);
	}

	bool BitmapFontProvider::HasGlyph(wchar_t code) const
	{
		if ( m_encoding.FromUCS2(code) != 0x1A ) return true; // TODO: replacement character constant
		return FontProvider::HasGlyph(code);
	}

	void BitmapFontProvider::PopulateTileset(GlyphTileset& tileset)
	{
		FontProvider::PopulateTileset(tileset);

		if ( m_cache.IsEmpty() )
		{
			LOG(Error, L"BitmapFontProvider has empty image cache");
			throw std::runtime_error("internal logic error");
		}

		if ( !m_cache.HasAlpha() ) m_cache.MakeTransparent(m_cache.GetPixel(Point(0, 0)));

		SizeF bitmap_size = SizeF(m_cache.GetSize().width, m_cache.GetSize().height);
		Size capacity
		(
			(size_t)std::floor(bitmap_size.width/m_glyph_size.width),
			(size_t)std::floor(bitmap_size.height/m_glyph_size.height)
		);

		for ( size_t i = 0; i < capacity.Area(); i++ )
		{
			size_t x = i%capacity.width;
			size_t y = (i-x)/capacity.width;
			std::uint16_t code = m_encoding.ToUCS2(int(i));

			Rectangle area = Rectangle(Point(x, y)*m_glyph_size, m_glyph_size);
			tileset.Add(code, m_cache, area);
		}

		m_cache.Dispose();
	}
}
