/*
* BearLibTerminal
* Copyright (C) 2013 Cfyz
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
* FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
* COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
* IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
* CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include <stdexcept>
#include <exception>
#include <fstream>
#include "DefaultFontProvider.hpp"
#include "Utility.hpp"
#include "Geometry.hpp"
#include "LoadBitmap.hpp"
#include "Resources.hpp"
#include "Log.hpp"

namespace BearLibTerminal
{
	DefaultFontProvider::DefaultFontProvider(const FontProvider::Options& options): FontProvider(options)
	{
		m_cell_size = m_glyph_size = Size(8, 16);
		m_cache = LoadBitmap(*Resources::OpenStream(L":DefaultFont"));
	}

	DefaultFontProvider::~DefaultFontProvider()
	{ }

	void DefaultFontProvider::Dispose() // FIXME: unnecessary
	{ }

	Bitmap DefaultFontProvider::GetGlyphBitmap(wchar_t code)
	{ 
		return FontProvider::GetGlyphBitmap(code);
	}

	bool DefaultFontProvider::HasGlyph(wchar_t code) const
	{
		/*
		0x0000..0x007F // 0080 // ASCII
		0x00A0..0x00FF // 0060 // Latin-1 supplement
		0x0370..0x03FF // 0090 // Greek and Coptic
		0x0400..0x04FF // 0100 // Cyrillic
		*/

		bool has_glyph = 
			(code >= 0x0000 && code <= 0x007F) ||
			(code >= 0x00A0 && code <= 0x00FF) ||
			(code >= 0x0370 && code <= 0x03FF) ||
			(code >= 0x0400 && code <= 0x04FF);

		return has_glyph? true: FontProvider::HasGlyph(code);
	}

	void DefaultFontProvider::PopulateTileset(GlyphTileset& tileset)
	{
		//LOG(Trace, L"Populating tileset via base FontProvider implementation");

		FontProvider::PopulateTileset(tileset);

		//LOG(Trace, L"Populating tileset with default font's glyphs");

		if ( m_cache.IsEmpty() )
		{
			LOG(Error, "[DefaultFontProvider::GetBasicGlyphs] Empty image cache");
			throw std::runtime_error("internal logic error");
		}

		if ( !m_cache.HasAlpha() ) m_cache.MakeTransparent(m_cache.GetPixel(Point(0, 0)));

		SizeF bitmap_size = SizeF(m_cache.GetSize().width, m_cache.GetSize().height);
		Size capacity
		(
			(size_t)std::floor(bitmap_size.width/m_glyph_size.width),
			(size_t)std::floor(bitmap_size.height/m_glyph_size.height)
		);

		struct range { size_t unicode_start, unicode_end; } ranges[] =
		{
			{ 0x0000, 0x007F },
			{ 0x00A0, 0x00FF },
			{ 0x0370, 0x03FF },
			{ 0x0400, 0x04FF }
		};
		const size_t N_ranges = sizeof(ranges)/sizeof(ranges[0]);

		std::list<std::pair<wchar_t, Bitmap>> result;

		for ( size_t i = 0, offset = 0; i < N_ranges; i++ )
		{
			size_t range_length = ranges[i].unicode_end - ranges[i].unicode_start + 1;

			for ( size_t j = 0; j < range_length; j++ )
			{
				size_t index = offset + j;
				size_t x = index%capacity.width;
				size_t y = (index-x)/capacity.width;
				wchar_t code = ranges[i].unicode_start + j;

				tileset.Add(code, m_cache, Rectangle(Point(x, y)*m_glyph_size, m_glyph_size));
			}

			offset += range_length;
		}

		m_cache.Dispose();
	}
}
