﻿/*
* BearLibTerminal
* Copyright (C) 2013 Cfyz
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
* FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
* COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
* IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
* CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include <stdexcept>
#include <cmath>
#include "FontProvider.hpp"
#include "Utility.hpp"
#include "Log.hpp"
#include "Geometry.hpp"
#include "LoadBitmap.hpp"
#include "Resources.hpp"
#include <stdlib.h>

namespace BearLibTerminal
{
	FontProvider::Options::Options()
	{ }

	bool FontProvider::Options::operator== (const Options& other) const
	{
		return
			(name == other.name) &&
			(size == other.size) &&
			(codepage == other.codepage) &&
			(mode == other.mode) &&
			(supplement == other.supplement);
	}

	bool FontProvider::Options::operator!= (const Options& other) const
	{
		return !(*this == other);
	}

	FontProvider::FontProvider(const Options& options)
	{
		// Deal with supplement
		if ( !options.supplement.empty() )
		{
			m_supplement_cache = LoadBitmap(*Resources::OpenStream(options.supplement));
		}
	}

	bool FontProvider::HasGlyph(wchar_t code) const
	{
		if ( code >= 0x2580 && code <= 0x259F ) return true;
		if ( code >= 0x2500 && code <= 0x257F ) return true;

		return false;
	}


	// TODO: remove all these unnecessary enums

	enum BlockElements // 2580-259F
	{
		kUpperHalfBlock									= 0x2580, // ▀
		kLowerOneEighthBlock							= 0x2581, // ▁
		kLowerOneQuarterBlock							= 0x2582, // ▂
		kLowerThreeEighthsBlock							= 0x2583, // ▃
		kLowerHalfBlock									= 0x2584, // ▄
		kLowerFiveEighthsBlock							= 0x2585, // ▅
		kLowerThreeQuartersBlock						= 0x2586, // ▆
		kLowerSevenEighthsBlock							= 0x2587, // ▇
		kFullBlock										= 0x2588, // █
		kLeftSevenEighthsBlock							= 0x2589, // ▉
		kLeftThreeQuartersBlock							= 0x258A, // ▊
		kLeftFiveEighthsBlock							= 0x258B, // ▋
		kLeftHalfBlock									= 0x258C, // ▌
		kLeftThreeEighthsBlock							= 0x258D, // ▍
		kLeftOneQuarterBlock							= 0x258E, // ▎
		kLeftOneEighthBlock								= 0x258F, // ▏
		kRightHalfBlock									= 0x2590, // ▐
		kLightShade										= 0x2591, // ░
		kMediumShade									= 0x2592, // ▒
		kDarkShade										= 0x2593, // ▓
		kUpperOneEighthBlock							= 0x2594, // ▔
		kRightOneEighthBlock							= 0x2595, // ▕
		kQuadrantLowerLeft								= 0x2596, // ▖
		kQuadrantLowerRight								= 0x2597, // ▗
		kQuadrantUpperLeft								= 0x2598, // ▘
		kQuadrantUpperLeftAndLowerLeftAndLowerRight		= 0x2599, // ▙
		kQuadrantUpperLeftAndLowerRight					= 0x259A, // ▚
		kQuadrantUpperLeftAndUpperRightAndLowerLeft		= 0x259B, // ▛
		kQuadrantUpperLeftAndUpperRightAndLowerRight	= 0x259C, // ▜
		kQuadrantUpperRight								= 0x259D, // ▝
		kQuadrantUpperRightAndLowerLeft					= 0x259E, // ▞
		kQuadrantUpperRightAndLowerLeftAndLowerRight	= 0x259F, // ▟
	};

	enum BoxDrawing // 2500-257F
	{
		kLightHorizontal								= 0x2500, // ─
		kHeavyHorisontal								= 0x2501, // ━
		kLightVertical									= 0x2502, // │
		kHeavyVertical									= 0x2503, // ┃
		kLightTripleDashHorisontal						= 0x2504, // ┄
		kHeavyTripleDashHorisontal						= 0x2505, // ┅
		kLightTripleDashVertical						= 0x2506, // ┆
		kHeavytTripleDashVertical						= 0x2507, // ┇
		kLightQuadrupleDashHorisontal					= 0x2508, // ┈
		kHeavyQuadrupleDashHorisontal					= 0x2509, // ┉
		kLightQuadrupleDashVertical						= 0x250A, // ┊
		kHeavyQuadrupleDashVertical						= 0x250B, // ┋
		kLightDownAndRight								= 0x250C, // ┌
		kDownLightAndRightHeavy							= 0x250D, // ┍
		kDownHeavyAndRightLight							= 0x250E, // ┎
		kHeavyDownAndRight								= 0x250F, // ┏
		kLightDownAndLeft								= 0x2510, // ┐
		kDownLightAndLeftHeavy							= 0x2511, // ┑
		kDownHeavyAndLeftLight							= 0x2512, // ┒
		kHeavyDownAndLeft								= 0x2513, // ┓
		kLightUpAndRight								= 0x2514, // └
		kUpLightAndRightHeavy							= 0x2515, // ┕
		kUpHeavyAndRightLight							= 0x2516, // ┖
		kHeavyUpAndRight								= 0x2517, // ┗
		kLightUpAndLeft									= 0x2518, // ┘
		kUpLightAndLeftHeavy							= 0x2519, // ┙
		kUpHeavyAndLeftLight							= 0x251a, // ┚
		kHeavyUpAndLeft									= 0x251b, // ┛
		kLightVerticalAndRight							= 0x251c, // ├
		kVerticalLightAndRightHeavy						= 0x251d, // ┝
		kUpHeavyAndRightDownLight						= 0x251e, // ┞
		kDownHeavyAndRightUpLight						= 0x251f, // ┟
		kVerticalHeavyAndRightLight						= 0x2520, // ┠
		kDownLightAndRightUpHeavy						= 0x2521, // ┡
		kUpLightAndRightDownHeavy						= 0x2522, // ┢
		kHeavyVerticalAndRight							= 0x2523, // ┣
		kLightVerticalAndLeft							= 0x2524, // ┤
		kVerticalLightAndLeftHeavy						= 0x2525, // ┥
		kUpHeavyAndLeftDownLight						= 0x2526, // ┦
		kDownHeavyAndLeftUpLight						= 0x2527, // ┧
		kVerticalHeavyAndLeftLight						= 0x2528, // ┨
		kDownLightAndLeftUpHeavy						= 0x2529, // ┩
		kUpLightAndLeftDownHeavy						= 0x252a, // ┪
		kHeavyVerticalAndLeft							= 0x252b, // ┫
		kLightDownAndHorisontal							= 0x252c, // ┬
		kLeftHeavyAndRightDownLight						= 0x252d, // ┭
		kRightHeavyAndLeftDownLight						= 0x252e, // ┮
		kDownLightAndHorisontalHeavy					= 0x252f, // ┯
		kDownHeavyAndHorisontalLight					= 0x2530, // ┰
		kRightLightAndLeftDownHeavy						= 0x2531, // ┱
		kLeftLightAndRightDownHeavy						= 0x2532, // ┲
		kHeavyDownAndHorisontal							= 0x2533, // ┳
		kLightUpAndHorisontal							= 0x2534, // ┴
		kLeftHeavyAndRightUpLight						= 0x2535, // ┵
		kRightHeavyAndLeftUpLight						= 0x2536, // ┶
		kUpLightAndHorisontalHeavy						= 0x2537, // ┷
		kUpHeavyAndHorisontalLight						= 0x2538, // ┸
		kRightLightAndLeftUpHeavy						= 0x2539, // ┹
		kLeftLightAndRightUpHeavy						= 0x253a, // ┺
		kHeavyUpAndHorisontal							= 0x253b, // ┻
		kLightVerticalAndHorisontal						= 0x253c, // ┼
		kLeftHeavyAndRightVerticalLight					= 0x253d, // ┽
		kRightHeavyAndLeftVerticalLight					= 0x253e, // ┾
		kVerticalLightAndHorisontalHeavy				= 0x253f, // ┿
		kUpHeavyAndDownHorisontalLight					= 0x2540, // ╀
		kDownHeavyAndUpHorisontalLight					= 0x2541, // ╁
		kVerticalHeavyAndHorisontalLight				= 0x2542, // ╂
		kLeftUpHeavyAndRightDownLight					= 0x2543, // ╃
		kRightUpHeavyAndLeftDownLight					= 0x2544, // ╄
		kLeftDownHeavyAndRightUpLight					= 0x2545, // ╅
		kRightDownHeavyAndLeftUpLight					= 0x2546, // ╆
		kDownLightAndUpHorisontalHeavy					= 0x2547, // ╇
		kUpLightAndDownHorisontalHeavy					= 0x2548, // ╈
		kRightLightAndLeftVerticalHeavy					= 0x2549, // ╉
		kLeftLightAndRightVerticalHeavy					= 0x254a, // ╊
		kHeavyVerticalAndHorisontal						= 0x254b, // ╋
		kLightDoubleDashHorisontal						= 0x254c, // ╌
		kHeavyDoubleDashHorisontal						= 0x254d, // ╍
		kLightDoubleDashVertical						= 0x254e, // ╎
		kHeavyDoubleDashVertical						= 0x254f, // ╏
		kDoubleHorisontal								= 0x2550, // ═
		kDoubleVertical									= 0x2551, // ║
		kDownSingleAndRightDouble						= 0x2552, // ╒
		kDownDoubleAndRightSingle						= 0x2553, // ╓
		kDoubleDownAndRight								= 0x2554, // ╔
		kDownSingleAndLeftDouble						= 0x2555, // ╕
		kDownDoubleAndLeftSingle						= 0x2556, // ╖
		kDoubleDownAndLeft								= 0x2557, // ╗
		kUpSingleAndRightDouble							= 0x2558, // ╘
		kUpDoubleAndRightSingle							= 0x2559, // ╙
		kDoubleUpAndRight								= 0x255a, // ╚
		kUpSingleAndLeftDouble							= 0x255b, // ╛
		kUpDoubleAndLeftSingle							= 0x255c, // ╜
		kDoubleUpAndLeft								= 0x255d, // ╝
		kVerticalSingleAndRightDouble					= 0x255e, // ╞
		kVerticalDoubleAndRightSingle					= 0x255f, // ╟
		kDoubleVerticalAndRight							= 0x2560, // ╠
		kVerticalSingleAndLeftDouble					= 0x2561, // ╡
		kVerticalDoubleAndLeftSingle					= 0x2562, // ╢
		kDoubleVerticalAndLeft							= 0x2563, // ╣
		kDownSingleAndHorisontalDouble					= 0x2564, // ╤
		kDownDoubleAndHorisontalSingle					= 0x2565, // ╥
		kDoubleDownAndHorisontal						= 0x2566, // ╦
		kUpSingleAndHorisontalDouble					= 0x2567, // ╧
		kUpDoubleAndHorisontalSingle					= 0x2568, // ╨
		kDoubleUpAndHorisontal							= 0x2569, // ╩
		kVerticalSingleAndHorisontalDouble				= 0x256a, // ╪
		kVerticalDoubleAndHorisontalSingle				= 0x256b, // ╫
		kDoubleVerticalAndHorisontal					= 0x256c, // ╬
		kLightArcDownAndRight							= 0x256d, // ╭
		kLightArcDownAndLeft							= 0x256e, // ╮
		kLightArcUpAndLeft								= 0x256f, // ╯
		kLightArcUpAndRight								= 0x2570, // ╰
		kLightDiagonalUpperRightToLowerLeft				= 0x2571, // ╱
		kLightDiagonalUpperLeftToLowerRight				= 0x2572, // ╲
		kLightDiagonalCross								= 0x2573, // ╳
		kLightLeft										= 0x2574, // ╴
		kLightUp										= 0x2575, // ╵
		kLightRight										= 0x2576, // ╶
		kLightDown										= 0x2577, // ╷
		kHeavyLeft										= 0x2578, // ╸
		kHeavyUp										= 0x2579, // ╹
		kHeavyRight										= 0x257a, // ╺
		kHeavyDown										= 0x257b, // ╻
		kLightLeftAndHeavyRight							= 0x257c, // ╼
		kLightUpAndHeavyDown							= 0x257d, // ╽
		kHeavyLeftAndLightRight							= 0x257e, // ╾
		kHeavyUpAndLightDown							= 0x257f  // ╿
	};

	enum LineType
	{
		kNone,
		kLight,
		kHeavy,
		kDouble
	};

	//
	// TODO: Rewrite, support lines wider than 1px, support arbitrary gap in double line
	//
	static void FillLines(Bitmap& bitmap, LineType left, LineType up, LineType right, LineType down)
	{
		Size size = bitmap.GetSize();
		int t = std::floor(size.width/7.0f); // Thickness
		if ( t == 0 ) t = 1; // Handle tiles <5 pixels width
		int cw = std::floor((size.width-1)/2.0f);
		int ch = std::floor((size.height-1)/2.0f);
		Color white(0xFFFFFFFF);
		int doubles = 0;

		const auto draw_vertical = [&](int y0, int y1, LineType type)
		{
			if ( type == kDouble ) doubles += 1;

			for ( int y = y0; y <= y1; y++ )
			{
				if ( type == kLight )
				{
					bitmap(cw, y) = white;
				}
				else if ( type == kHeavy )
				{
					bitmap(cw-1, y) = white;
					bitmap(cw+0, y) = white;
					bitmap(cw+1, y) = white;
				}
				else if ( type == kDouble )
				{
					bitmap(cw-1, y) = white;
					bitmap(cw+1, y) = white;
				}
			}
		};

		const auto draw_horisontal = [&](int x0, int x1, LineType type)
		{
			if ( type == kDouble ) doubles += 1;

			for ( int x = x0; x <= x1; x++ )
			{
				if ( type == kLight )
				{
					bitmap(x, ch) = white;
				}
				else if ( type == kHeavy )
				{
					bitmap(x, ch-1) = white;
					bitmap(x, ch+0) = white;
					bitmap(x, ch+1) = white;
				}
				else if ( type == kDouble )
				{
					bitmap(x, ch-1) = white;
					bitmap(x, ch+1) = white;
				}
			}
		};

		if ( left != kNone ) draw_horisontal(0, cw , left);
		if ( up != kNone ) draw_vertical(0, ch, up);
		if ( right != kNone ) draw_horisontal(cw, size.width-1, right);
		if ( down != kNone ) draw_vertical(ch, size.height-1, down);

		// Some corner cases (pun intended):

		// Missing pixel on heavy lines
		if ( left == kHeavy && up == kHeavy && right != kHeavy && down != kHeavy ) bitmap(cw+1, ch+1) = white;
		if ( left != kHeavy && up == kHeavy && right == kHeavy && down != kHeavy ) bitmap(cw-1, ch+1) = white;
		if ( left != kHeavy && up != kHeavy && right == kHeavy && down == kHeavy ) bitmap(cw-1, ch-1) = white;
		if ( left == kHeavy && up != kHeavy && right != kHeavy && down == kHeavy ) bitmap(cw+1, ch-1) = white;
		
		// Missing pixel on double lines
		if ( left == kDouble && up == kDouble && right != kDouble && down != kDouble ) bitmap(cw+1, ch+1) = white;
		if ( left != kDouble && up == kDouble && right == kDouble && down != kDouble ) bitmap(cw-1, ch+1) = white;
		if ( left != kDouble && up != kDouble && right == kDouble && down == kDouble ) bitmap(cw-1, ch-1) = white;
		if ( left == kDouble && up != kDouble && right != kDouble && down == kDouble ) bitmap(cw+1, ch-1) = white;

		// Clogged double lines
		if ( doubles >= 2 )
		{
			Color transparent(0x00000000);
			if ( left == kDouble ) for ( int x = 0; x <= cw; x++ ) bitmap(x, ch) = transparent;
			if ( up == kDouble ) for ( int y = 0; y <= ch; y++ ) bitmap(cw, y) = transparent;
			if ( right == kDouble ) for ( int x = cw; x < size.width; x++ ) bitmap(x, ch) = transparent;
			if ( down == kDouble ) for ( int y = ch; y < size.height; y++ ) bitmap(cw, y) = transparent;
		}
	}

	//
	// TODO: Rewrite, makes poor gaps
	//
	static void FillDash(Bitmap& bitmap, bool vertical, bool heavy, int n)
	{
		Size size = bitmap.GetSize();
		int t = std::floor(size.width/7.0f); // Thickness
		if ( t == 0 ) t = 1; // Handle tiles <5 pixels width
		int w = heavy? t*3: t;
		int d = std::floor(w/2.0f);
		int cw = std::floor((size.width-1)/2.0f);
		int ch = std::floor((size.height-1)/2.0f);
		Color white(0xFFFFFFFF);

		// NOTE: rudimentary support of arbitrary line width
		const auto hslice = [&](int y, std::function<void(Color&)> f)
		{
			for ( int x = cw-d; x < cw-d+w; x++ ) f(bitmap(x, y));
		};

		const auto vslice = [&](int x, std::function<void(Color&)> f)
		{
			for ( int y = ch-d; y < ch-d+w; y++ ) f(bitmap(x, y));
		};

		if ( vertical )
		{
			// Solid line
			for ( int y = 0; y < size.height; y++ ) hslice(y, [&white](Color& c) { c = white; } );

			float T = size.height / (float)n;
			float Td = 0.25f / size.height;
			float q = size.height/(float)n/2.0f;
			if ( q > 1 ) q = std::floor(q);
			if ( q < 1 ) q = 1.0f; else if ( q > t ) q = t;

			for ( int i = 0; i < n; i++ )
			{
				float b = (i+1)*T-Td, a = (i+1)*T-q;
				int ai = std::floor(a), bi = std::floor(b);
				float af = a-ai, bf = bi+1-b;

				for ( int j = ai; j <= bi; j++ ) hslice(j, [](Color& c) { c = Color(0x00FFFFFF); });
				hslice(ai, [&af](Color& c) { c.a += af*255; });
				hslice(bi, [&bf](Color& c) { c.a += bf*255; });
			}
		}
		else
		{
			for ( int x = 0; x < size.width; x++ ) vslice(x, [&white](Color& c) { c = white; } );

			float T = size.width / (float)n;
			float Td = 0.25f / size.width;
			float q = size.width/(float)n/2.0f;
			if ( q > 1 ) q = std::floor(q); 
			if ( q > t ) q = t;

			for ( int i = 0; i < n; i++ )
			{
				float b = (i+1)*T-Td, a = (i+1)*T-q;
				int ai = std::floor(a), bi = std::floor(b);
				float af = a-ai, bf = bi+1-b;

				for ( int j = ai; j <= bi; j++ ) vslice(j, [](Color& c) { c = Color(0x00FFFFFF); });
				vslice(ai, [&af](Color& c) { c.a += af*255; });
				vslice(bi, [&bf](Color& c) { c.a += bf*255; });
			}
		}
	}

	//
	// TODO: Rewrite this, support arbitrary width and antialias
	//
	static void FillQuarterCircle(Bitmap& bitmap, int quadrant)
	{
		Size size = bitmap.GetSize();
		int t = std::floor(size.width/7.0f); // Thickness
		if ( t == 0 ) t = 1; // Handle tiles <5 pixels width
		int cw = std::floor((size.width-1)/2.0f);
		int ch = std::floor((size.height-1)/2.0f);

		auto putpixel = [&](int x, int y)
		{
			if ( x >= 0 && y >= 0 && x < size.width && y < size.height )
			{
				bitmap(x, y) = Color(0xFFFFFFFF);
			}
		};

		int rx = cw, ry = ch;
		int x0 = 0, y0 = 0;
		int sx = 1, sy = 1;

		if ( quadrant == 1 || quadrant == 2 )
		{
			ry = size.height-1-ch;
			y0 = size.height-1;
			sy = -1;
		}

		if ( quadrant == 3 || quadrant == 2 )
		{
			rx = size.width-1-cw;
			x0 = size.width-1;
			sx = -1;
		}

		int rxsq=rx*rx;
		int rysq=ry*ry;
		int tworxsq=2*rxsq;
		int tworysq=2*rysq;
		int x=0;
		int y=ry;
		int d1=rysq - (rxsq * ry) + (0.25 * rxsq);
		int dx= tworysq * x;
		int dy= tworxsq * y;

		do
		{
			putpixel(x0+x*sx, y0+y*sy);

			if (d1 < 0)
			{
				x=x+1;
				//y=y;
				dx=dx + tworysq;
				d1=d1 + dx + rysq;
			}
			else
			{
				x=x+1;
				y=y-1;
				dx= dx + tworysq;
				dy= dy - tworxsq;
				d1= d1 + dx - dy + rysq;
			}
		}
		while (dx < dy);
		
		int d2 = rysq * ( x + 0.5) * ( x + 0.5 ) + rxsq * (y - 1) * (y-1) - rxsq * rysq;
		do
		{
			putpixel(x0+x*sx, y0+y*sy);

			if (d2 >0)
			{
				//x=x;
				y=y-1;
				dy = dy - tworxsq;
				d2 = d2 - dy + rxsq;
			}
			else
			{
				x= x+1;
				y=y-1;
				dy=dy - tworxsq;
				dx= dx + tworysq;
				d2 = d2 + dx -dy + rxsq;
			}
		}
		while ( y >= 0 );
	}

	//
	// TODO: Rewrite this, support arbitrary width and antialias
	//
	static void FillDiagonal(Bitmap& bitmap, bool forward, bool backward)
	{
		Size size = bitmap.GetSize();
		Color white(0xFFFFFFFF);

		auto draw_line = [&](int x0, int y0, int x1, int y1)
		{
			/*
			int dx = x1-x0;
			int dy = y1-y0;
			x0 -= dx; x1 += dx;
			y0 -= dy; y1 += dy;
			dx = abs(x1-x0);
			dy = abs(y1-y0);
			/*/
			int dx = abs(x1-x0);
			int dy = abs(y1-y0);
			//*/
			int sx = (x0 < x1)? 1: -1;
			int sy = (y0 < y1)? 1: -1;
			int err = dx-dy;
 
			while ( true )
			{
				//if ( x0 >= 0 && y0 >= 0 && x0 < size.width && y0 < size.height ) 
					bitmap(x0, y0) = white;
				if ( x0 == x1 && y0 == y1 ) break;
				int e2 = 2*err;
				if ( e2 > -dy )
				{
					err = err - dy;
					x0 = x0 + sx;
				}
				if ( e2 < dx )
				{
					err = err + dx;
					y0 = y0 + sy;
				}
			}
		};

		if ( forward ) draw_line(0, size.height-1, size.width-1, 0);
		if ( backward ) draw_line(0, 0, size.width-1, size.height-1);
	}

	static void FillFullBlock(Bitmap& bitmap, float opacity)
	{
		Color c(255, 255, 255, 255 * opacity);
		for_each_point(Rectangle(bitmap.GetSize()), [&](Point pt) { bitmap.SetPixel(pt, c); });
	}

	static void FillHorisontal(Bitmap& bitmap, float top, float bottom)
	{
		Size size = bitmap.GetSize();
		top *= size.height;
		bottom *= size.height;
		/* // Turning off anti-aliasing, TODO: option
		top = std::floor(top+0.5f);
		bottom = std::floor(bottom+0.5f);
		//*/
		int ftop = std::floor(top);
		int cbottom = std::ceil(bottom);
		for ( int y = ftop; y < cbottom; y++ )
		{
			float a = 1.0f;
			if ( y == ftop ) a = (ftop+1-top);
			if ( y == cbottom-1 ) a = bottom-(cbottom-1);
			Color c(255, 255, 255, 255*a);
			for ( int x = 0; x < size.width; x++ ) bitmap.SetPixel(Point(x, y), c);
		}
	}

	static void FillVertical(Bitmap& bitmap, float left, float right)
	{
		Size size = bitmap.GetSize();
		left *= size.width;
		right *= size.width;
		/* // Turning off anti-aliasing, TODO: option
		left = std::floor(left+0.5f);
		right = std::floor(right+0.5f);
		//*/
		int fleft = std::floor(left);
		int cright = std::ceil(right);
		for ( int x = fleft; x < cright; x++ )
		{
			float a = 1.0f;
			if ( x == fleft ) a = (fleft+1-left);
			if ( x == cright-1 ) a = right-(cright-1);
			Color c(255, 255, 255, 255*a);
			for ( int y = 0; y < size.height; y++ ) bitmap.SetPixel(Point(x, y), c);
		}
	}

	// TODO: refactor
	static void FillQuadrant(Bitmap& bitmap, bool topleft, bool topright, bool bottomright, bool bottomleft)
	{
		Size size = bitmap.GetSize();
		//size.height -= 1;
		float cw = size.width/2.0f;
		float ch = size.height/2.0f;
		int fcw = std::floor(cw), ccw = std::ceil(cw);
		int fch = std::floor(ch), cch = std::ceil(ch);

		for_each_point(Rectangle(size), [&](Point p){ bitmap(p) = 0x00FFFFFF; });

		if ( topleft )
		{
			for_each_point(Rectangle(0, 0, fcw, fch), [&](Point p){ bitmap(p).a = 255; });
			if ( cw > fcw )
			{
				uint8_t a = 255*(cw-fcw);
				for ( int y = 0; y < fch; y++ )
				{
					bitmap(Point(fcw, y)).a += a;
				}
			}
			if ( ch > fch )
			{
				uint8_t a = 255*(ch-fch);
				for ( int x = 0; x < fcw; x++ )
				{
					bitmap(Point(x, fch)).a += a;
				}
			}
			if ( ch > fch && cw > fcw )
			{
				uint8_t a = 255*(((cw-fcw)+(ch-fch))/4.0f);
				bitmap(Point(fcw, fch)).a += a;
			}
		}

		if ( topright )
		{
			for_each_point(Rectangle(ccw, 0, size.width-ccw, fch), [&](Point p){ bitmap(p).a = 255; });
			if ( ccw > cw )
			{
				uint8_t a = 255*(ccw-cw);
				for ( int y = 0; y < fch; y++ )
				{
					bitmap(Point(ccw-1, y)).a += a;
				}
			}
			if ( ch > fch )
			{
				uint8_t a = 255*(ch-fch);
				for ( int x = ccw; x < size.width; x++ )
				{
					bitmap(Point(x, fch)).a += a;
				}
			}
			if ( ch > fch && ccw > cw )
			{
				uint8_t a = 255*(((ccw-cw)+(ch-fch))/4.0f);
				bitmap(Point(ccw-1, fch)).a += a;
			}
		}

		if ( bottomright )
		{
			for_each_point(Rectangle(ccw, cch, size.width-ccw, size.height-cch), [&](Point p){ bitmap(p).a = 255; });
			if ( ccw > cw )
			{
				uint8_t a = 255*(ccw-cw);
				for ( int y = cch; y < size.height; y++ )
				{
					bitmap(Point(ccw-1, y)).a += a;
				}
			}
			if ( cch > ch )
			{
				uint8_t a = 255*(cch-ch);
				for ( int x = ccw; x < size.width; x++ )
				{
					bitmap(Point(x, cch-1)).a += a;
				}
			}
			if ( cch > ch && ccw > cw )
			{
				uint8_t a = 255*(((ccw-cw)+(cch-ch))/4.0f);
				bitmap(Point(ccw-1, cch-1)).a += a;
			}
		}

		if ( bottomleft )
		{
			for_each_point(Rectangle(0, cch, fcw, size.height-cch), [&](Point p){ bitmap(p).a = 255; });
			if ( cw > fcw )
			{
				uint8_t a = 255*(cw-fcw);
				for ( int y = cch; y < size.height; y++ )
				{
					bitmap(Point(fcw, y)).a += a;
				}
			}
			if ( cch > ch )
			{
				uint8_t a = 255*(cch-ch);
				for ( int x = 0; x < fcw; x++ )
				{
					bitmap(Point(x, cch-1)).a += a;
				}
			}
			if ( cch > ch && cw > fcw )
			{
				uint8_t a = 255*(((cw-fcw)+(cch-ch))/4.0f);
				bitmap(Point(fcw, cch-1)).a += a;
			}
		}
	}

	Bitmap FontProvider::GetGlyphBitmap(wchar_t code)
	{
		Bitmap result(m_glyph_size, Color(0x00000000));
		Bitmap canvas(m_cell_size, Color(0x00000000));
		Color white(0xFFFFFFFF);

		switch ( code )
		{
		// Blocks
		case kUpperHalfBlock:						FillHorisontal(canvas, 0.0f, 0.5f); break; 
		case kLowerOneEighthBlock:					FillHorisontal(canvas, 0.875f, 1.0f); break;
		case kLowerOneQuarterBlock:					FillHorisontal(canvas, 0.75f, 1.0f); break;
		case kLowerThreeEighthsBlock:				FillHorisontal(canvas, 0.625f, 1.0f); break;
		case kLowerHalfBlock:						FillHorisontal(canvas, 0.5f, 1.0f); break;
		case kLowerFiveEighthsBlock:				FillHorisontal(canvas, 0.375f, 1.0f); break;
		case kLowerThreeQuartersBlock:				FillHorisontal(canvas, 0.25f, 1.0f); break;
		case kLowerSevenEighthsBlock:				FillHorisontal(canvas, 0.125f, 1.0f); break;
		case kFullBlock:							FillFullBlock(canvas, 1.0f); break;
		case kLeftSevenEighthsBlock:				FillVertical(canvas, 0.0f, 0.875f); break;
		case kLeftThreeQuartersBlock:				FillVertical(canvas, 0.0f, 0.75f); break;
		case kLeftFiveEighthsBlock:					FillVertical(canvas, 0.0f, 0.625f); break;
		case kLeftHalfBlock:						FillVertical(canvas, 0.0f, 0.5f); break;
		case kLeftThreeEighthsBlock:				FillVertical(canvas, 0.0f, 0.375f); break;
		case kLeftOneQuarterBlock:					FillVertical(canvas, 0.0f, 0.25f); break;
		case kLeftOneEighthBlock:					FillVertical(canvas, 0.0f, 0.125f); break;
		case kRightHalfBlock:						FillVertical(canvas, 0.5f, 1.0f); break;
		case kLightShade:							FillFullBlock(canvas, 0.25f); break;
		case kMediumShade:							FillFullBlock(canvas, 0.5f); break;
		case kDarkShade:							FillFullBlock(canvas, 0.75f); break;
		case kUpperOneEighthBlock:					FillHorisontal(canvas, 0.0f, 0.125f); break;
		case kRightOneEighthBlock:					FillVertical(canvas, 0.875f, 1.0f); break;

		// Quadrants
		case kQuadrantLowerLeft:							FillQuadrant(canvas, false, false, false, true); break;
		case kQuadrantLowerRight:							FillQuadrant(canvas, false, false, true, false); break;
		case kQuadrantUpperLeft:							FillQuadrant(canvas, true, false, false, false); break;
		case kQuadrantUpperLeftAndLowerLeftAndLowerRight:	FillQuadrant(canvas, true, false, true, true); break;
		case kQuadrantUpperLeftAndLowerRight:				FillQuadrant(canvas, true, false, true, false); break;
		case kQuadrantUpperLeftAndUpperRightAndLowerLeft:	FillQuadrant(canvas, true, true, false, true); break;
		case kQuadrantUpperLeftAndUpperRightAndLowerRight:	FillQuadrant(canvas, true, true, true, false); break;
		case kQuadrantUpperRight:							FillQuadrant(canvas, false, true, false, false); break;
		case kQuadrantUpperRightAndLowerLeft:				FillQuadrant(canvas, false, true, false, true); break;
		case kQuadrantUpperRightAndLowerLeftAndLowerRight:	FillQuadrant(canvas, false, true, true, true); break;

		// Full lines
		case kLightHorizontal:						FillLines(canvas, kLight, kNone, kLight, kNone); break;
		case kHeavyHorisontal:						FillLines(canvas, kHeavy, kNone, kHeavy, kNone); break;
		case kLightVertical:						FillLines(canvas, kNone, kLight, kNone, kLight); break;
		case kHeavyVertical:						FillLines(canvas, kNone, kHeavy, kNone, kHeavy); break;
		
		// Triple and quadruple dashes
		case kLightTripleDashHorisontal:			FillDash(canvas, false, false, 3); break;
		case kHeavyTripleDashHorisontal:			FillDash(canvas, false, true, 3); break;
		case kLightTripleDashVertical:				FillDash(canvas, true, false, 3); break;
		case kHeavytTripleDashVertical:				FillDash(canvas, true, true, 3); break;
		case kLightQuadrupleDashHorisontal:			FillDash(canvas, false, false, 4); break;
		case kHeavyQuadrupleDashHorisontal:			FillDash(canvas, false, true, 4); break;
		case kLightQuadrupleDashVertical:			FillDash(canvas, true, false, 4); break;
		case kHeavyQuadrupleDashVertical:			FillDash(canvas, true, true, 4); break;
		
		// Corners (two lines)
		case kLightDownAndRight:					FillLines(canvas, kNone, kNone, kLight, kLight); break;
		case kDownLightAndRightHeavy:				FillLines(canvas, kNone, kNone, kHeavy, kLight); break;
		case kDownHeavyAndRightLight:				FillLines(canvas, kNone, kNone, kLight, kHeavy); break;
		case kHeavyDownAndRight:					FillLines(canvas, kNone, kNone, kHeavy, kHeavy); break;
		case kLightDownAndLeft:						FillLines(canvas, kLight, kNone, kNone, kLight); break;
		case kDownLightAndLeftHeavy:				FillLines(canvas, kHeavy, kNone, kNone, kLight); break;
		case kDownHeavyAndLeftLight:				FillLines(canvas, kLight, kNone, kNone, kHeavy); break;
		case kHeavyDownAndLeft:						FillLines(canvas, kHeavy, kNone, kNone, kHeavy); break;
		case kLightUpAndRight:						FillLines(canvas, kNone, kLight, kLight, kNone); break;
		case kUpLightAndRightHeavy:					FillLines(canvas, kNone, kLight, kHeavy, kNone); break;
		case kUpHeavyAndRightLight:					FillLines(canvas, kNone, kHeavy, kLight, kNone); break;
		case kHeavyUpAndRight:						FillLines(canvas, kNone, kHeavy, kHeavy, kNone); break;
		case kLightUpAndLeft:						FillLines(canvas, kLight, kLight, kNone, kNone); break;
		case kUpLightAndLeftHeavy:					FillLines(canvas, kHeavy, kLight, kNone, kNone); break;
		case kUpHeavyAndLeftLight:					FillLines(canvas, kLight, kHeavy, kNone, kNone); break;
		case kHeavyUpAndLeft:						FillLines(canvas, kHeavy, kHeavy, kNone, kNone); break;

		// Triangles (three lines)
		case kLightVerticalAndRight:				FillLines(canvas, kNone, kLight, kLight, kLight); break;
		case kVerticalLightAndRightHeavy:			FillLines(canvas, kNone, kLight, kHeavy, kLight); break;
		case kUpHeavyAndRightDownLight:				FillLines(canvas, kNone, kHeavy, kLight, kLight); break;
		case kDownHeavyAndRightUpLight:				FillLines(canvas, kNone, kLight, kLight, kHeavy); break;
		case kVerticalHeavyAndRightLight:			FillLines(canvas, kNone, kHeavy, kLight, kHeavy); break;
		case kDownLightAndRightUpHeavy:				FillLines(canvas, kNone, kHeavy, kHeavy, kLight); break;
		case kUpLightAndRightDownHeavy:				FillLines(canvas, kNone, kLight, kHeavy, kHeavy); break;
		case kHeavyVerticalAndRight:				FillLines(canvas, kNone, kHeavy, kHeavy, kHeavy); break;
		case kLightVerticalAndLeft:					FillLines(canvas, kLight, kLight, kNone, kLight); break;
		case kVerticalLightAndLeftHeavy:			FillLines(canvas, kHeavy, kLight, kNone, kLight); break;
		case kUpHeavyAndLeftDownLight:				FillLines(canvas, kLight, kHeavy, kNone, kLight); break;
		case kDownHeavyAndLeftUpLight:				FillLines(canvas, kLight, kLight, kNone, kHeavy); break;
		case kVerticalHeavyAndLeftLight:			FillLines(canvas, kLight, kHeavy, kNone, kHeavy); break;
		case kDownLightAndLeftUpHeavy:				FillLines(canvas, kHeavy, kHeavy, kNone, kLight); break;
		case kUpLightAndLeftDownHeavy:				FillLines(canvas, kHeavy, kLight, kNone, kHeavy); break;
		case kHeavyVerticalAndLeft:					FillLines(canvas, kHeavy, kHeavy, kNone, kHeavy); break;

		// Triangles (three lines), cont.
		case kLightDownAndHorisontal:				FillLines(canvas, kLight, kNone, kLight, kLight); break;
		case kLeftHeavyAndRightDownLight:			FillLines(canvas, kHeavy, kNone, kLight, kLight); break;
		case kRightHeavyAndLeftDownLight:			FillLines(canvas, kLight, kNone, kHeavy, kLight); break;
		case kDownLightAndHorisontalHeavy:			FillLines(canvas, kHeavy, kNone, kHeavy, kLight); break;
		case kDownHeavyAndHorisontalLight:			FillLines(canvas, kLight, kNone, kLight, kHeavy); break;
		case kRightLightAndLeftDownHeavy:			FillLines(canvas, kHeavy, kNone, kLight, kHeavy); break;
		case kLeftLightAndRightDownHeavy:			FillLines(canvas, kLight, kNone, kHeavy, kHeavy); break;
		case kHeavyDownAndHorisontal:				FillLines(canvas, kHeavy, kNone, kHeavy, kHeavy); break;
		case kLightUpAndHorisontal:					FillLines(canvas, kLight, kLight, kLight, kNone); break;
		case kLeftHeavyAndRightUpLight:				FillLines(canvas, kHeavy, kLight, kLight, kNone); break;
		case kRightHeavyAndLeftUpLight:				FillLines(canvas, kLight, kLight, kHeavy, kNone); break;
		case kUpLightAndHorisontalHeavy:			FillLines(canvas, kHeavy, kLight, kHeavy, kNone); break;
		case kUpHeavyAndHorisontalLight:			FillLines(canvas, kLight, kHeavy, kLight, kNone); break;
		case kRightLightAndLeftUpHeavy:				FillLines(canvas, kHeavy, kHeavy, kLight, kNone); break;
		case kLeftLightAndRightUpHeavy:				FillLines(canvas, kLight, kHeavy, kHeavy, kNone); break;
		case kHeavyUpAndHorisontal:					FillLines(canvas, kHeavy, kHeavy, kHeavy, kNone); break;

		// Crosses (four lines)
		case kLightVerticalAndHorisontal:			FillLines(canvas, kLight, kLight, kLight, kLight); break;
		case kLeftHeavyAndRightVerticalLight:		FillLines(canvas, kHeavy, kLight, kLight, kLight); break;
		case kRightHeavyAndLeftVerticalLight:		FillLines(canvas, kLight, kLight, kHeavy, kLight); break;
		case kVerticalLightAndHorisontalHeavy:		FillLines(canvas, kHeavy, kLight, kHeavy, kLight); break;
		case kUpHeavyAndDownHorisontalLight:		FillLines(canvas, kLight, kHeavy, kLight, kLight); break;
		case kDownHeavyAndUpHorisontalLight:		FillLines(canvas, kLight, kLight, kLight, kHeavy); break;
		case kVerticalHeavyAndHorisontalLight:		FillLines(canvas, kLight, kHeavy, kLight, kHeavy); break;
		case kLeftUpHeavyAndRightDownLight:			FillLines(canvas, kHeavy, kHeavy, kLight, kLight); break;
		case kRightUpHeavyAndLeftDownLight:			FillLines(canvas, kLight, kHeavy, kHeavy, kLight); break;
		case kLeftDownHeavyAndRightUpLight:			FillLines(canvas, kHeavy, kLight, kLight, kHeavy); break;
		case kRightDownHeavyAndLeftUpLight:			FillLines(canvas, kLight, kLight, kHeavy, kHeavy); break;
		case kDownLightAndUpHorisontalHeavy:		FillLines(canvas, kHeavy, kHeavy, kHeavy, kLight); break;
		case kUpLightAndDownHorisontalHeavy:		FillLines(canvas, kHeavy, kLight, kHeavy, kHeavy); break;
		case kRightLightAndLeftVerticalHeavy:		FillLines(canvas, kHeavy, kHeavy, kLight, kHeavy); break;
		case kLeftLightAndRightVerticalHeavy:		FillLines(canvas, kLight, kHeavy, kHeavy, kHeavy); break;
		case kHeavyVerticalAndHorisontal:			FillLines(canvas, kHeavy, kHeavy, kHeavy, kHeavy); break;
		
		// Double dashes
		case kLightDoubleDashHorisontal:			FillDash(canvas, false, false, 2); break;
		case kHeavyDoubleDashHorisontal:			FillDash(canvas, false, true, 2); break;
		case kLightDoubleDashVertical:				FillDash(canvas, true, false, 2); break;
		case kHeavyDoubleDashVertical:				FillDash(canvas, true, true, 2); break;
		
		// Double corners (two lines)
		case kDoubleHorisontal:						FillLines(canvas, kDouble, kNone, kDouble, kNone); break;
		case kDoubleVertical:						FillLines(canvas, kNone, kDouble, kNone, kDouble); break;
		case kDownSingleAndRightDouble:				FillLines(canvas, kNone, kNone, kDouble, kLight); break;
		case kDownDoubleAndRightSingle:				FillLines(canvas, kNone, kNone, kLight, kDouble); break;
		case kDoubleDownAndRight:					FillLines(canvas, kNone, kNone, kDouble, kDouble); break;
		case kDownSingleAndLeftDouble:				FillLines(canvas, kDouble, kNone, kNone, kLight); break;
		case kDownDoubleAndLeftSingle:				FillLines(canvas, kLight, kNone, kNone, kDouble); break;
		case kDoubleDownAndLeft:					FillLines(canvas, kDouble, kNone, kNone, kDouble); break;
		case kUpSingleAndRightDouble:				FillLines(canvas, kNone, kLight, kDouble, kNone); break;
		case kUpDoubleAndRightSingle:				FillLines(canvas, kNone, kDouble, kLight, kNone); break;
		case kDoubleUpAndRight:						FillLines(canvas, kNone, kDouble, kDouble, kNone); break;
		case kUpSingleAndLeftDouble:				FillLines(canvas, kDouble, kLight, kNone, kNone); break;
		case kUpDoubleAndLeftSingle:				FillLines(canvas, kLight, kDouble, kNone, kNone); break;
		case kDoubleUpAndLeft:						FillLines(canvas, kDouble, kDouble, kNone, kNone); break;
		
		// Double triangles and crosses (three and four lines)
		case kVerticalSingleAndRightDouble:			FillLines(canvas, kNone, kLight, kDouble, kLight); break;
		case kVerticalDoubleAndRightSingle:			FillLines(canvas, kNone, kDouble, kLight, kDouble); break;
		case kDoubleVerticalAndRight:				FillLines(canvas, kNone, kDouble, kDouble, kDouble); break;
		case kVerticalSingleAndLeftDouble:			FillLines(canvas, kDouble, kLight, kNone, kLight); break;
		case kVerticalDoubleAndLeftSingle:			FillLines(canvas, kLight, kDouble, kNone, kDouble); break;
		case kDoubleVerticalAndLeft:				FillLines(canvas, kDouble, kDouble, kNone, kDouble); break;
		case kDownSingleAndHorisontalDouble:		FillLines(canvas, kDouble, kNone, kDouble, kLight); break;
		case kDownDoubleAndHorisontalSingle:		FillLines(canvas, kLight, kNone, kLight, kDouble); break;
		case kDoubleDownAndHorisontal:				FillLines(canvas, kDouble, kNone, kDouble, kDouble); break;
		case kUpSingleAndHorisontalDouble:			FillLines(canvas, kDouble, kLight, kDouble, kNone); break;
		case kUpDoubleAndHorisontalSingle:			FillLines(canvas, kLight, kDouble, kLight, kNone); break;
		case kDoubleUpAndHorisontal:				FillLines(canvas, kDouble, kDouble, kDouble, kNone); break;
		case kVerticalSingleAndHorisontalDouble:	FillLines(canvas, kDouble, kLight, kDouble, kLight); break;
		case kVerticalDoubleAndHorisontalSingle:	FillLines(canvas, kLight, kDouble, kLight, kDouble); break;
		case kDoubleVerticalAndHorisontal:			FillLines(canvas, kDouble, kDouble, kDouble, kDouble); break;
		
		// Arcs (elliptic quadrants)
		case kLightArcDownAndRight:					FillQuarterCircle(canvas, 2); break;
		case kLightArcDownAndLeft:					FillQuarterCircle(canvas, 1); break;
		case kLightArcUpAndLeft:					FillQuarterCircle(canvas, 4); break;
		case kLightArcUpAndRight:					FillQuarterCircle(canvas, 3); break;
		
		// Diagonals
		case kLightDiagonalUpperRightToLowerLeft:	FillDiagonal(canvas, true, false); break;
		case kLightDiagonalUpperLeftToLowerRight:	FillDiagonal(canvas, false, true); break;
		case kLightDiagonalCross:					FillDiagonal(canvas, true, true); break;
		
		// Single and mixed lines
		case kLightLeft:							FillLines(canvas, kLight, kNone, kNone, kNone); break;
		case kLightUp:								FillLines(canvas, kNone, kLight, kNone, kNone); break;
		case kLightRight:							FillLines(canvas, kNone, kNone, kLight, kNone); break;
		case kLightDown:							FillLines(canvas, kNone, kNone, kNone, kLight); break;
		case kHeavyLeft:							FillLines(canvas, kHeavy, kNone, kNone, kNone); break;
		case kHeavyUp:								FillLines(canvas, kNone, kHeavy, kNone, kNone); break;
		case kHeavyRight:							FillLines(canvas, kNone, kNone, kHeavy, kNone); break;
		case kHeavyDown:							FillLines(canvas, kNone, kNone, kNone, kHeavy); break;
		case kLightLeftAndHeavyRight:				FillLines(canvas, kLight, kNone, kHeavy, kNone); break;
		case kLightUpAndHeavyDown:					FillLines(canvas, kNone, kLight, kNone, kHeavy); break;
		case kHeavyLeftAndLightRight:				FillLines(canvas, kHeavy, kNone, kLight, kNone); break;
		case kHeavyUpAndLightDown:					FillLines(canvas, kNone, kHeavy, kNone, kLight); break;

		default:
			{
				for ( int x=1; x<m_cell_size.width-1; x++ )
				{
					canvas.SetPixel(Point(x, 1), white);
					canvas.SetPixel(Point(x, m_cell_size.height-2), white);
				}

				for ( int y=1; y<m_cell_size.height-1; y++ )
				{
					canvas.SetPixel(Point(1, y), white);
					canvas.SetPixel(Point(m_cell_size.width-2, y), white);
				}
			}
			break;
		}

		int x_offset = (m_glyph_size.width-m_cell_size.width)/2;
		result.Blit(canvas, Point(x_offset, 0));
		return result;
	}
	
	void FontProvider::PopulateTileset(GlyphTileset& tileset)
	{
		LOG(Trace, L"Populating tileset with most basic glyphs (base FontProvider implementation)");
		tileset.Add(0xFFFF, GetGlyphBitmap(0xFFFF));

		Bitmap& cache = m_supplement_cache;
		if ( !cache.IsEmpty() )
		{
			// TODO: populate tileset with supplement
			if ( !cache.HasAlpha() ) cache.MakeTransparent(cache.GetPixel(Point(0, 0)));

			SizeF bitmap_size = SizeF(cache.GetSize().width, cache.GetSize().height);
			Size capacity
			(
				(size_t)std::floor(bitmap_size.width/m_glyph_size.width),
				(size_t)std::floor(bitmap_size.height/m_glyph_size.height)
			);
			LOG(Info, "Supplement font image has " << capacity.width << "x" << capacity.height << " glyphs");

			for ( int i = 0; i < capacity.Area(); i++ )
			{
				size_t x = i%capacity.width;
				size_t y = (i-x)/capacity.width;
				std::uint16_t code = 0xE000 + i;

				Rectangle area = Rectangle(Point(x, y)*m_glyph_size, m_glyph_size);
				tileset.Add(code, cache, area);
			}
		}
	}

	void FontProvider::PopulateTileset(GlyphTileset& tileset, const std::list<std::uint16_t>& codes)
	{
		for ( wchar_t code: codes )
		{
			tileset.Add(code, GetGlyphBitmap(HasGlyph(code)? code: 0xFFFF));
		}
	}
}
