/*
* BearLibTerminal
* Copyright (C) 2013 Cfyz
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
* FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
* COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
* IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
* CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include "FreetypeFontProvider.hpp"
#include <freetype/ftlcdfil.h>
#include <freetype/ftglyph.h>
#include "Utility.hpp"
#include "Encoding.hpp"

namespace BearLibTerminal
{
	Bitmap FreetypeFontProvider::PrepareBitmap(FT_GlyphSlot slot)
	{
		int rows = slot->bitmap.rows;
		int columns = 0;
		int pixel_size = 0;

		int height = m_font_face->size->metrics.height >> 6;
		int descender = m_font_face->size->metrics.descender >> 6;
		int bx = slot->metrics.horiBearingX >> 6;
		int by = slot->metrics.horiBearingY >> 6;

		if ( slot->bitmap.pixel_mode == FT_PIXEL_MODE_GRAY )
		{
			columns = slot->bitmap.width;
			pixel_size = 1;
		}
		else if ( slot->bitmap.pixel_mode == FT_PIXEL_MODE_LCD )
		{
			columns = slot->bitmap.width/3;
			pixel_size = 3;
		}
		else if ( slot->bitmap.pixel_mode == FT_PIXEL_MODE_MONO )
		{
			columns = slot->bitmap.width;
			pixel_size = 0;
		}

		// Read bitmap from Freetype glyph slot
		Bitmap glyph(Size(columns, rows), Color(0x00000000));
		
		for ( int y = 0; y < rows; y++ )
		{
			for ( int x = 0; x < columns; x++ )
			{
				uint8_t* p = slot->bitmap.buffer + y*slot->bitmap.pitch + x*pixel_size;
				if ( slot->bitmap.pixel_mode == FT_PIXEL_MODE_GRAY )
				{
					Color c(255, 255, 255, p[0]);
					glyph.SetPixel(Point(x, y), c);
				}
				else if ( slot->bitmap.pixel_mode == FT_PIXEL_MODE_LCD )
				{
					Color c(p[2], p[1], p[0], 255);
					glyph.SetPixel(Point(x, y), c);
				}
				else if ( slot->bitmap.pixel_mode == FT_PIXEL_MODE_MONO )
				{
					int j = x%8;
					int i = (x-j)/8;
					uint8_t byte = *(slot->bitmap.buffer + y*slot->bitmap.pitch + i);
					uint8_t alpha = (byte & (1 << (7-j)))? 255: 0;
					glyph(Point(x, y)) = Color(255, 255, 255, alpha);
				}
			}
		}

		// Gamma correction (there are fonts which are rendered somewhat grey)
		// NOTE: correction math differs slightly between normal and lcd (alpha handled differently)
		if ( slot->bitmap.pixel_mode == FT_PIXEL_MODE_GRAY )
		{
			int max = 0;

			auto f1 = [&](Point p)
			{
				Color& c = glyph(p);
				int t = c.a;
				if ( t > max ) max = t;
			};
			for_each_point(Rectangle(glyph.GetSize()), f1);

			float gamma_correction = 255.0f / max;

			if ( gamma_correction > 1.0f )
			{
				auto f2 = [&](Point p)
				{
					Color& c = glyph(p);
					c.a = std::min<float>(255, c.a*gamma_correction);
				};
				for_each_point(Rectangle(glyph.GetSize()), f2);
			}
		}
		else if ( slot->bitmap.pixel_mode == FT_PIXEL_MODE_LCD )
		{
			int max = 0;

			auto f = [&](Point p)
			{
				Color& c = glyph(p);
				int t = (c.r + c.g + c.b)/3;
				if ( t > max ) max = t;
			};
			for_each_point(Rectangle(glyph.GetSize()), f);

			float gamma_correction = 255.0f / max;

			if ( gamma_correction > 1.0f )
			{
				auto f2 = [&](Point p)
				{
					Color& c = glyph(p);
					c.r = std::min<float>(255, c.r*gamma_correction);
					c.g = std::min<float>(255, c.g*gamma_correction);
					c.b = std::min<float>(255, c.b*gamma_correction);
				};
				for_each_point(Rectangle(glyph.GetSize()), f2);
			}
		}

		int offset = 0;

		if ( slot->bitmap.pixel_mode == FT_PIXEL_MODE_LCD )
		{
			offset = -1;
		}

		//int suboffset = bx > 0? std::floor(bx/64.0): -std::floor(-bx/64.0);
		int x = offset + std::floor(std::abs(bx)/64.0) * (bx >= 0? 1: -1);
		int y = height+descender-by;

		Bitmap result(m_glyph_size, Color(0x00000000));
		result.BlitUnchecked(glyph, BasicPoint<int>(x, y));

		return result;
	}

	/*
	FreetypeFontProvider::FreetypeFontProvider(dictionary_t& options):
		m_font_library(nullptr),
		m_font_face(nullptr),
		m_render_mode(FT_RENDER_MODE_NORMAL)
	{
		// type, filename, size, mode

		if ( options.count(L"filename") == 0 )
		{
			throw std::runtime_error("[FreetypeFontProvider] Filename parameter is missing");
		}

		std::wstring filename = options[L"filename"];
		float size = 14.0f;
		
		if ( options.count(L"size") == 0 || !try_parse(options[L"size"], size) || size < 1 || size > 72 )
		{
			throw std::runtime_error("[FreetypeFontProvider] Size parameter is missing or invalid");
		}

		if ( options.count(L"mode") )
		{
			std::wstring mode = options[L"mode"];
			if ( mode == L"normal" ) m_render_mode = FT_RENDER_MODE_NORMAL;
			else if ( mode == L"monochrome" ) m_render_mode = FT_RENDER_MODE_MONO;
			else if ( mode == L"lcd" ) m_render_mode = FT_RENDER_MODE_LCD;
			else throw std::runtime_error("[FreetypeFontProvider] Render mode parameter is invalid");
		}

		if ( FT_Init_FreeType(&m_font_library) )
		{
			throw std::runtime_error("[FreetypeFontProvider] Can't initialize Freetype");
		}

		std::string filename_u8 = Encoding::UTF8.FromUCS2(filename);
		if ( FT_New_Face(m_font_library, filename_u8.c_str(), 0, &m_font_face) )
		{
			Dispose();
			throw std::runtime_error("[FreetypeFontProvider] Can't load font from file");
		}

		int hres = 64;
		FT_Matrix matrix =
		{
			(int)((1.0/hres) * 0x10000L),
			(int)((0.0)      * 0x10000L),
			(int)((0.0)      * 0x10000L),
			(int)((1.0)      * 0x10000L)
		};

		if ( FT_Set_Char_Size(m_font_face, (uint32_t)(size*64), 0, 96*hres, 96) )
		{
			Dispose();
			throw std::runtime_error("[FreetypeFontProvider] Can't setup font");
		}

		FT_Set_Transform(m_font_face, &matrix, NULL);

		float yscale = m_font_face->size->metrics.y_scale / (float)(1 << 16);
		int height = std::ceil((m_font_face->bbox.yMax-m_font_face->bbox.yMin) * yscale / 64.0f);
		int descender = std::floor(m_font_face->descender * yscale / 64.0f);
		int ascender = std::ceil(m_font_face->ascender * yscale / 64.0f);

		auto get_metrics = [&](wchar_t code) -> FT_Glyph_Metrics
		{
			if ( FT_Load_Glyph(m_font_face, FT_Get_Char_Index(m_font_face, code), 0) )
			{
				throw std::runtime_error("[FreetypeFontProvider] Metrics error");
			}

			return m_font_face->glyph->metrics;
		};

		int width = 0;

		FT_Glyph_Metrics mDot = get_metrics(L'.'), mAt = get_metrics(L'@');
		if ( mDot.horiAdvance == mAt.horiAdvance )
		{
			// Monospace font
			width = std::ceil(mDot.horiAdvance / 64.0f);
		}
		else
		{
			// This is not a monospace font
			//throw std::runtime_error("[FreetypeFontProvider::ctor] This is not a monospace font");
			width = std::ceil(mAt.horiAdvance / 64.0f);
		}

		width = std::ceil(width/64.0);

		height = m_font_face->size->metrics.height >> 6;
		descender = m_font_face->size->metrics.descender >> 6;

		m_cell_size = Size(width, height);
		m_glyph_size = Size(width, height);

		FT_Library_SetLcdFilter(m_font_library, FT_LCD_FILTER_DEFAULT);
		
		return;
	}
	*/

	FreetypeFontProvider::FreetypeFontProvider(const FontProvider::Options& options): FontProvider(options)
	{
		if ( options.size.height < 1 )
		{
			throw std::runtime_error("[FreetypeFontProvider] Invalid size");
		}

		if ( options.mode == L"" || options.mode == L"normal" )
		{
			m_render_mode = FT_RENDER_MODE_NORMAL;
		}
		else if ( options.mode == L"monochrome" )
		{
			m_render_mode = FT_RENDER_MODE_MONO;
		}
		else if ( options.mode == L"lcd" )
		{
			m_render_mode = FT_RENDER_MODE_LCD;
		}
		else
		{
			throw std::runtime_error("[FreetypeFontProvider] Invalid font.mode");
		}

		if ( FT_Init_FreeType(&m_font_library) )
		{
			throw std::runtime_error("[FreetypeFontProvider] Can't initialize Freetype");
		}

		std::string filename_u8 = Encoding::UTF8.FromUCS2(options.name);
		if ( FT_New_Face(m_font_library, filename_u8.c_str(), 0, &m_font_face) )
		{
			Dispose();
			throw std::runtime_error("[FreetypeFontProvider] Can't load font from file");
		}

		int hres = 64;
		FT_Matrix matrix =
		{
			(int)((1.0/hres) * 0x10000L),
			(int)((0.0)      * 0x10000L),
			(int)((0.0)      * 0x10000L),
			(int)((1.0)      * 0x10000L)
		};

		if ( FT_Set_Char_Size(m_font_face, (uint32_t)(options.size.height*64), 0, 96*hres, 96) )
		{
			Dispose();
			throw std::runtime_error("[FreetypeFontProvider] Can't setup font");
		}

		FT_Set_Transform(m_font_face, &matrix, NULL);

		float yscale = m_font_face->size->metrics.y_scale / (float)(1 << 16);
		int height = std::ceil((m_font_face->bbox.yMax-m_font_face->bbox.yMin) * yscale / 64.0f);
		//int descender = std::floor(m_font_face->descender * yscale / 64.0f);
		//int ascender = std::ceil(m_font_face->ascender * yscale / 64.0f);

		auto get_metrics = [&](wchar_t code) -> FT_Glyph_Metrics
		{
			if ( FT_Load_Glyph(m_font_face, FT_Get_Char_Index(m_font_face, code), 0) )
			{
				throw std::runtime_error("[FreetypeFontProvider] Metrics error");
			}

			return m_font_face->glyph->metrics;
		};

		int width = 0;

		FT_Glyph_Metrics mDot = get_metrics(L'.'), mAt = get_metrics(L'@');
		if ( mDot.horiAdvance == mAt.horiAdvance )
		{
			// Monospace font
			width = std::ceil(mDot.horiAdvance / 64.0f);
		}
		else
		{
			// This is not a monospace font
			//throw std::runtime_error("[FreetypeFontProvider::ctor] This is not a monospace font");
			width = std::ceil(mAt.horiAdvance / 64.0f);
		}

		width = std::ceil(width/64.0);

		height = m_font_face->size->metrics.height >> 6;
		//descender = m_font_face->size->metrics.descender >> 6;

		m_cell_size = Size(width, height);
		m_glyph_size = Size(width, height);

		FT_Library_SetLcdFilter(m_font_library, FT_LCD_FILTER_DEFAULT);
		
		return;
	}

	FreetypeFontProvider::~FreetypeFontProvider()
	{
		Dispose();
	}

	void FreetypeFontProvider::Dispose()
	{
		// NYI
	}

	Bitmap FreetypeFontProvider::GetGlyphBitmap(wchar_t code)
	{
		unsigned int index = 0;
		int rc = 0;

		if ( FontProvider::HasGlyph(code) || (index = FT_Get_Char_Index(m_font_face, code)) == 0 )
		{
			return FontProvider::GetGlyphBitmap(code);
		}
		
		if ( FT_Load_Glyph(m_font_face, index, /*m_render_mode == FT_RENDER_MODE_MONO? 0:*/ FT_LOAD_FORCE_AUTOHINT) )
		{
			throw std::runtime_error("[FreetypeFontProvider::GetGlyphBitmap] Can't load character glyph");
		}

		if ( m_font_face->glyph->format != FT_GLYPH_FORMAT_BITMAP )
		{
			FT_Render_Mode render_mode = m_render_mode;

			if ( (rc = FT_Render_Glyph(m_font_face->glyph, render_mode)) != 0 )
			{
				//ShowMessageBox(to_string(rc));
				//FT_Err_Unknown_File_Format;
				throw std::runtime_error("[FreetypeFontProvider::GetGlyphBitmap] Can't render glyph");
			}
		}

		return PrepareBitmap(m_font_face->glyph);
	}

	bool FreetypeFontProvider::HasGlyph(wchar_t code) const
	{
		if ( FT_Get_Char_Index(m_font_face, code) > 0 ) return true;
		return FontProvider::HasGlyph(code);
	}

	void FreetypeFontProvider::PopulateTileset(GlyphTileset& tileset, const std::list<wchar_t>& codes)
	{
		std::list<std::uint16_t> not_found;

		for ( wchar_t code: codes )
		{
			if ( HasGlyph(code) )
			{
				tileset.Add(code, GetGlyphBitmap(code));
			}
			else
			{
				not_found.push_back(code);
			}
		}

		FontProvider::PopulateTileset(tileset, not_found);
	}
}
