/*
* BearLibTerminal
* Copyright (C) 2013 Cfyz
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
* FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
* COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
* IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
* CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include "Log.hpp"
#include "Encoding.hpp"
#include "Utility.hpp"

#ifdef _WIN32
#define WIN32_LEAN_AND_MEAN
#include <Windows.h>
#else
#include <fstream>
#include <sys/time.h>
#endif

#include <iostream>

namespace BearLibTerminal
{
	Log::Options::Options():
		file(L"bearlibterminal.log"),
		level(Log::Level::Error)
	{ }

	Log::Options::Options(const std::wstring& file, Log::Level level):
		file(file),
		level(level)
	{ }

	std::mutex Log::m_lock;
	Log::Options Log::m_options;

	/*
	static void TruncateFile(const std::wstring& filename)
	{
		if ( filename.empty() ) return;

		HANDLE h = CreateFileW
		(
			filename.c_str(),
			GENERIC_WRITE,
			FILE_SHARE_READ|FILE_SHARE_WRITE,
			NULL,
			TRUNCATE_EXISTING,
			FILE_ATTRIBUTE_NORMAL,
			NULL
		);

		if ( h != NULL ) CloseHandle(h);
	}
	*/

#ifdef _WIN32
	void Log::Write(Log::Level level, const std::wstring& what)
	{
		if ( level > m_options.level ) return;

		//*
		std::wstring filename = GetLocked(m_options.file, m_lock);
		if ( filename.empty() ) return;
		/*/
		std::wstring filename;

		// Synchronously retrieve filename and truncate the file if necessary
		{
			std::lock_guard<std::mutex> guard(m_lock);
			filename = m_options.filename;
			if ( filename.empty() ) return;
			if ( m_options.truncate && !m_file_truncated )
			{
				TruncateFile(filename);
				m_file_truncated = true;
			}
		}
		//*/

		HANDLE h = CreateFileW
		(
			filename.c_str(),
			GENERIC_WRITE,
			FILE_SHARE_READ|FILE_SHARE_WRITE,
			NULL,
			OPEN_ALWAYS,
			FILE_ATTRIBUTE_NORMAL,
			NULL
		);

		if ( h == NULL ) return;

		SetFilePointer(h, 0, 0, FILE_END);

		// Prepare time
		SYSTEMTIME systime;
		GetSystemTime(&systime);
		char buffer[13]; // hh:mm:ss.ttt
		_snprintf
		(
			buffer, 13,
			"%02d:%02d:%02d.%03d",
			systime.wHour,
			systime.wMinute,
			systime.wSecond,
			systime.wMilliseconds
		);

		std::ostringstream stream;
		//stream << "[" << buffer << "] " << Encoding::UTF8.FromUCS2(what) << "\r\n";
		std::string level_str = Encoding::UTF8.FromUCS2(to_wstring(level));
		stream << buffer << " [Terminal " << level_str << "] " << Encoding::UTF8.FromUCS2(what) << "\r\n";
		std::string what_u8 = stream.str();

		DWORD written;
		WriteFile(h, what_u8.c_str(), what_u8.length(), &written, NULL);

		CloseHandle(h);
	}
#else
	void Log::Write(Log::Level level, const std::wstring& what)
	{
		if ( level > m_options.level ) return;

		std::wstring filename = GetLocked(m_options.file, m_lock);
		if ( filename.empty() ) return;

		std::ofstream file(Encoding::UTF8.FromUCS2(filename).c_str(), std::ios_base::out|std::ios_base::app);

		timeval tv;
		gettimeofday(&tv, NULL);
		struct tm tm;
		localtime_r(&tv.tv_sec, &tm);

		char buffer[13]; // hh:mm:ss.ttt
		snprintf
		(
			buffer, 13,
			"%02d:%02d:%02d.%03d",
			tm.tm_hour,
			tm.tm_min,
			tm.tm_sec,
			(int)(tv.tv_usec/1000)
		);

		file << "[" << buffer << "] " << Encoding::UTF8.FromUCS2(what) << "\n";
	}
#endif

	void Log::SetOptions(const Log::Options& options)
	{
		std::lock_guard<std::mutex> guard(m_lock);
		m_options = options;
	}

	Log::Level Log::GetLevel()
	{
		return m_options.level;
	}

	std::wostream& operator<< (std::wostream& stream, const Log::Level& value)
	{
		stream << [=]() -> std::wstring
		{
			switch ( value )
			{
				case Log::Level::None: return L"None";
				case Log::Level::Fatal: return L"Fatal";
				case Log::Level::Error: return L"Error";
				case Log::Level::Warning: return L"Warning";
				case Log::Level::Info: return L"Info";
				case Log::Level::Debug: return L"Debug";
				case Log::Level::Trace: return L"Trace";
			}
			return to_wstring((int)value);
		}();
		return stream;
	}

	std::wistream& operator>> (std::wistream& stream, Log::Level& value)
	{
		auto str = extract<std::wstring>(stream);
		if ( str == L"None" ) value = Log::Level::None;
		else if ( str == L"Fatal" ) value = Log::Level::Fatal;
		else if ( str == L"Error" ) value = Log::Level::Error;
		else if ( str == L"Warning" ) value = Log::Level::Warning;
		else if ( str == L"Info" ) value = Log::Level::Info;
		else if ( str == L"Debug" ) value = Log::Level::Debug;
		else if ( str == L"Trace" ) value = Log::Level::Trace;
		else stream.setstate(std::ios::failbit);
		return stream;
	}
}

#ifdef _WIN32
std::wstring GetLastErrorStr(std::uint32_t rc)
{
	const size_t buffer_size = 256;
	wchar_t buffer[buffer_size];
	DWORD n = FormatMessageW(FORMAT_MESSAGE_FROM_SYSTEM, 0, rc, MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_US), buffer, buffer_size, NULL);
	std::wstring result(buffer, n);
	if ( result.find(L"\r\n") != std::wstring::npos ) result = result.substr(0, result.length()-2);
	if ( result.length() > 0 && result.back() == '.' ) result = result.substr(0, result.length()-1);
	return result;
}

std::wstring GetLastErrorStr()
{
	return GetLastErrorStr(GetLastError());
}
#endif
