/*
* BearLibTerminal
* Copyright (C) 2013 Cfyz
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
* FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
* COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
* IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
* CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include "Options.hpp"
#include "Utility.hpp"
#include "Log.hpp"

namespace BearLibTerminal
{
	Timeout::Timeout(): value(-1) // Indefinite
	{ }

	Timeout::Timeout(int value): value(value)
	{ }

	Timeout::operator int() const
	{
		return value;
	}

	Timeout Timeout::Indefinite(-1);
	Timeout Timeout::Immediate(0);

	std::wostream& operator<<(std::wostream& stream, const Timeout& value)
	{
		if ( (int)value < 0 )
		{
			stream << L"infinite";
		}
		else if ( value == Timeout::Immediate )
		{
			stream << L"immediate";
		}
		else
		{
			stream << (int)value;
		}

		return stream;
	}

	std::wistream& operator>>(std::wistream& stream, Timeout& value)
	{
		std::wstring str = extract<std::wstring>(stream);

		if ( str == L"infinite" )
		{
			value = Timeout::Indefinite;
		}
		else if ( str == L"immediate" || str == L"nonblocked" )
		{
			value = Timeout::Immediate;
		}
		else
		{
			int temp;
			if ( try_parse<int>(str, temp) )
			{
				value = temp;
			}
			else
			{
				stream.setstate(std::ios::failbit);
			}
		}

		return stream;
	}

	InputFilter::InputFilter(): value(0)
	{ }

	InputFilter::InputFilter(unsigned int value): value(value)
	{ }

	InputFilter::operator unsigned int() const
	{
		return value;
	}

	bool InputFilter::Has(unsigned int other) const
	{
		return (value & other) != 0;
	}

	InputFilter InputFilter::None(0);
	InputFilter InputFilter::KeyPress(0x01);
	InputFilter InputFilter::KeyRelease(0x02);
	InputFilter InputFilter::MouseMove(0x04);
	InputFilter InputFilter::MouseScroll(0x08);
	InputFilter InputFilter::All(0x0F);

	static struct { const std::wstring s; InputFilter v; } flag_list[] =
	{
		{ L"all", InputFilter::All },
		{ L"keypress", InputFilter::KeyPress },
		{ L"keyrelease", InputFilter::KeyRelease },
		{ L"mousemove", InputFilter::MouseMove },
		{ L"mousescroll", InputFilter::MouseScroll },
		{ L"none", InputFilter::None }
	};

	static const size_t N_flags = sizeof(flag_list)/sizeof(flag_list[0]);

	std::wostream& operator<<(std::wostream& stream, const InputFilter& value)
	{
		bool first = true;
		for ( size_t i = 0; i < N_flags; i++ )
		{
			if ( value & flag_list[i].v )
			{
				if ( !first ) stream << L"+";
				stream << flag_list[i].s;
				first = false;
			}
		}

		return stream;
	}

	std::wistream& operator>>(std::wistream& stream, InputFilter& value)
	{
		std::wstring str = extract<std::wstring>(stream);
		unsigned int flags = 0;

		for ( size_t i = 0; i < N_flags; i++ )
		{
			size_t n = str.find(flag_list[i].s);
			if ( n != std::wstring::npos )
			{
				unsigned int value = flag_list[i].v;
				if ( n > 0 && str[n-1] == L'-' ) flags &= ~value; else flags |= value;
			}
		}

		value = flags;
		return stream;
	}

	Options::Options():
		size(80, 25),
		title(L"BearLibTerminal"),
		icon(L"app.ico"),
		output_postformat(true),
		output_synchronous(true),
		output_custom_render(nullptr),
		log(L"bearlibterminal.log", Log::Level::Error),
		//input_events(InputEvents::KeyPress),
		input_events(InputFilter::KeyPress),
		input_precise_mouse(false),
		input_sticky_close(true),
		input_timeout(-1)
	{ }
}
