/*
* BearLibTerminal
* Copyright (C) 2013 Cfyz
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
* FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
* COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
* IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
* CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include "SpriteContainer.hpp"
#include "Resources.hpp"
#include "LoadBitmap.hpp"
#include "Utility.hpp"
#include "Log.hpp"
#include "OpenGL.hpp"
#include <cstring>

namespace BearLibTerminal
{
	SpriteContainer::SpriteContainer()
	{
		std::memset(m_flags, 0, sizeof(m_flags));
	}

	SpriteContainer::~SpriteContainer()
	{
		Clear();
	}

	void SpriteContainer::Clear()
	{
		std::memset(m_flags, 0, sizeof(m_flags));
		m_sprites.clear();
	}

	bool SpriteContainer::Add(std::uint16_t code, const std::wstring& filename)
	{
		if (code < 0 || code > 0xFFFF) return false;

		try
		{
			std::shared_ptr<std::istream> stream = Resources::OpenStream(filename);
			Bitmap bitmap = LoadBitmap(*stream);

			Size bitmap_size = bitmap.GetSize();
			Size texture_size = Size(pow2roundup(bitmap_size.width), pow2roundup(bitmap_size.height));

			Bitmap canvas = Bitmap(texture_size, Color(0xFF804000));
			canvas.Blit(bitmap, Point());
			std::unique_ptr<Texture> texture(new Texture(canvas));

			Sprite& i = m_sprites[code];
			i.texture = std::move(texture);
			i.bitmap_size = bitmap_size;
			i.texture_size = texture_size;
			m_flags[code] = 1;
			return true;
		}
		catch (std::exception& e)
		{
			LOG(Warning, "Exception caught while loading new image glyphs: " << e.what());
			return false;
		}
	}

	void SpriteContainer::Remove(std::uint16_t code)
	{
		if (code < 0 || code > 0xFFFF || !m_flags[code]) return;
		m_sprites.erase(code);
		m_flags[code] = 0;
	}

	bool SpriteContainer::HasSprite(std::uint16_t code)
	{
		if (code < 0 || code > 0xFFFF) return 0;
		return m_flags[code];
	}

	void SpriteContainer::Draw(std::uint16_t code, Point location, Size cell_size)
	{
		if (code < 0 || code > 0xFFFF || !m_flags[code]) return;
		Sprite& i = m_sprites[code];
		if (i.texture)
		{
			int sprite_left = location.x * cell_size.width;
			int sprite_top = location.y * cell_size.height;
			int sprite_width = (int)std::floor(i.bitmap_size.width/(float)cell_size.width)*cell_size.width;
			int sprite_height = (int)std::floor(i.bitmap_size.height/(float)cell_size.height)*cell_size.height;
			float tu = sprite_width/(float)i.texture_size.width;
			float tv = sprite_height/(float)i.texture_size.height;

			i.texture->Bind();
			glBegin(GL_QUADS);
			// Top-left
			glTexCoord2f(0.0f, 0.0f);
			glVertex2i(sprite_left, sprite_top);
			// Bottom-left
			glTexCoord2f(0.0f, tv);
			glVertex2i(sprite_left, sprite_top+sprite_height);
			// Bottom-right
			glTexCoord2f(tu, tv);
			glVertex2i(sprite_left+sprite_width, sprite_top+sprite_height);
			// Top-right
			glTexCoord2f(tu, 0.0f);
			glVertex2i(sprite_left+sprite_width, sprite_top);
			glEnd();
		}
	}
}
