/*
* BearLibTerminal
* Copyright (C) 2013 Cfyz
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
* FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
* COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
* IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
* CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include "Terminal.hpp"
#include <limits>
#include "Utility.hpp"
#include "Geometry.hpp"
#include "Log.hpp"
#include "FontProviderFactory.hpp"
#include "OpenGL.hpp"
#include "Palette.hpp"
#include "Resources.hpp"
#include "LoadBitmap.hpp"
#include <string>
#include <cstring>

namespace BearLibTerminal
{
	Terminal::Terminal(int codepage):
		m_state(0x100, 0),
		m_status(kUninitialized),
		m_flag_custom_render(false),
		m_flag_enforced_redraw(false)
	{
		m_encoding = Encoding(codepage);
		m_scene.backbuffer.resize(m_options.size.Area());
		m_scene.frontbuffer = m_scene.backbuffer;

		m_window = Window::Create();

		m_window->SetOnDestroy(std::bind(&Terminal::OnWindowClose, this));
		m_window->SetOnInput(std::bind(&Terminal::OnWindowInput, this, std::placeholders::_1));
		m_window->SetOnRedraw(std::bind(&Terminal::OnWindowRedraw, this));
		m_window->SetOnActivate(std::bind(&Terminal::OnWindowActivate, this));

		m_scene.font = FontProviderFactory::Create(m_options.font);
		m_scene.font->PopulateTileset(m_scene.tileset);
		m_state[VK_CELL_WIDTH] = m_scene.font->GetCellSize().width;
		m_state[VK_CELL_HEIGHT] = m_scene.font->GetCellSize().height;

		m_window->SetTitle(m_options.title);
		m_window->SetIcon(m_options.icon);
		m_window->SetClientSize(m_options.size*m_scene.font->GetCellSize());
	}

	Terminal::~Terminal()
	{
		// NOTE: Window object must be deconstructed before Terminal object
		// because window will invoke OnWindowClose callback
		m_window.reset();
	}

	int Terminal::SetOptions(const char* options)
	{
		try
		{
			std::wstring options_str = m_encoding.ToUCS2(options);
			LOG(Debug, "Trying to set new options: " << options_str);

			auto options = parse_dictionary(options_str);

			std::unique_lock<std::mutex> guard(m_lock);
			Options temp = m_options;

			bool update_size = false;
			bool update_title = false;
			bool update_icon = false;
			bool update_font = false;
			bool update_log = false;
			std::shared_ptr<FontProvider> font_provider;

			if ( options.count(L"size") )
			{
				if ( !try_parse(options[L"size"], temp.size) )
				{
					throw std::invalid_argument("terminal size is not valid");
				}

				if ( temp.size.Area() == 0 || !Rectangle(Size(256, 256)).Contains(Rectangle(temp.size)) )
				{
					throw std::invalid_argument("terminal size is not valid");
				}

				update_size = true;
			}

			if ( options.count(L"title") )
			{
				temp.title = options[L"title"];
				update_title = true;
			}

			if ( options.count(L"icon") )
			{
				const std::wstring& filename = options[L"icon"];
				if ( !m_window->ValidateIcon(filename) ) // TODO: no validation yet
				{
					throw std::runtime_error("invalid icon file");
				}
				temp.icon = filename;
				update_icon = true;
			}

			if ( options.count(L"font.name") )
			{
				temp.font.name = options[L"font.name"];
				update_font = true;
			}

			if ( options.count(L"font.size") )
			{
				if ( options[L"font.size"].find(L"x") != std::wstring::npos ) // Most likely, "WxH" for bitmap font
				{
					if ( !try_parse(options[L"font.size"], temp.font.size) || temp.font.size.Area() == 0 )
					{
						throw std::runtime_error("font size is not valid");
					}
				}
				else // Most likely "H" for freetype font
				{
					temp.font.size.width = 0;

					if ( !try_parse(options[L"font.size"], temp.font.size.height) || temp.font.size.height == 0 )
					{
						throw std::runtime_error("font size is not valid");
					}
				}

				update_font = true;
			}

			if ( options.count(L"font.codepage") )
			{
				temp.font.codepage = options[L"font.codepage"];
				update_font = true;
			}

			if ( options.count(L"font.mode") )
			{
				temp.font.mode = options[L"font.mode"];
				update_font = true;
			}

			if ( options.count(L"font.supplement") )
			{
				temp.font.supplement = options[L"font.supplement"];
				update_font = true;
			}

			if ( update_font )
			{
				if ( temp.font != m_options.font )
				{
					// Allocate new font provider, MAY FAIL (and it's ok)
					font_provider = FontProviderFactory::Create(temp.font);
				}
				else
				{
					update_font = false;
				}
			}

			if ( options.count(L"input.precise_mousemove") )
			{
				temp.input_precise_mouse = parse<bool>(options[L"input.precise_mousemove"]);
			}

			if ( options.count(L"input.events") )
			{
				temp.input_events = parse<InputFilter>(options[L"input.events"]);
			}

			if ( options.count(L"input.timeout") )
			{
				if ( !try_parse<Timeout>(options[L"input.timeout"], temp.input_timeout) )
				{
					throw std::runtime_error("input timeout is not valid");
				}
			}

			if ( options.count(L"output.postformat") )
			{
				temp.output_postformat = parse<bool>(options[L"output.postformat"]);
			}

			if ( options.count(L"log.file") )
			{
				temp.log.file = options[L"log.file"];
				update_log = true;
			}

			if ( options.count(L"log.level") )
			{
				if ( !try_parse<Log::Level>(options[L"log.level"], temp.log.level) )
				{
					throw std::runtime_error("log level is not valid");
				}

				update_log = true;
			}

			// Apply options (SHOULD NOT fail from here on)
			LOG(Trace, "Options have been parsed, applying");
			m_options = temp;

			bool refresh_output = false;

			if ( update_size )
			{
				LOG(Trace, "Setting new window size " << m_options.size);
				m_scene.backbuffer = std::vector<Cell>(m_options.size.Area());
				m_scene.frontbuffer = m_scene.backbuffer;
				m_window->SetClientSize(m_options.size*m_scene.font->GetGlyphSize());
				m_scene.initialized = false;
				refresh_output = true;
			}

			if ( update_title )
			{
				LOG(Trace, "Setting new window title \"" << m_options.title << "\"");
				m_window->SetTitle(m_options.title);
			}

			if ( update_icon )
			{
				LOG(Trace, "Setting new window icon from file \"" << m_options.icon << "\"");
				m_window->SetIcon(m_options.icon);
			}

			if ( update_font && font_provider )
			{
				// Apply new (already allocated and therefore checked) font provider
				LOG(Trace, "Setting new font provider");
				m_scene.tileset.Clear();
				m_scene.font = font_provider;
				m_scene.font->PopulateTileset(m_scene.tileset);
				m_state[VK_CELL_WIDTH] = m_scene.font->GetCellSize().width;
				m_state[VK_CELL_HEIGHT] = m_scene.font->GetCellSize().height;
				m_window->SetClientSize(m_options.size*m_scene.font->GetCellSize());
				m_scene.initialized = false;
				refresh_output = true;
			}

			if ( update_log )
			{
				Log::SetOptions(m_options.log);
			}

			guard.unlock(); // Done with synchronization

			if ( refresh_output && m_status == kNormal )
			{
				Refresh();
			}

			LOG(Info, L"The new options were successfully set");

			return 0;
		}
		catch(std::exception& e)
		{
			LOG(Error, L"Setting options failed: " << e.what());
			return -1;
		}
	}

	void Terminal::Refresh()
	{
		// Refresh is completely ignored in custom-render mode
		if (m_flag_custom_render) return;

		if ( m_status == kUninitialized )
		{
			m_window->Show();
			m_status = kNormal;
			m_scene.initialized = false;
		}

		// Ignore irrelevant redraw calls
		if (m_status != kNormal) return;

		{
			std::lock_guard<std::mutex> guard(m_lock);
			m_scene.frontbuffer = m_scene.backbuffer;
		}

		// NOTE: this call will syncronously wait OnWindowRedraw completion
		ScopedSet<volatile bool> enforced_redraw_guard(m_flag_enforced_redraw, true);
		m_window->ForceRedraw();
	}

	void Terminal::Clear()
	{
		std::lock_guard<std::mutex> guard(m_lock); // Call modifies the scene

		for ( auto& i: m_scene.backbuffer )
		{
			i.tiles.clear();
			i.background = m_scene.background;
		}
	}

	void Terminal::Clear(int x, int y, int w, int h)
	{
		std::lock_guard<std::mutex> guard(m_lock); // Call modifies the scene

		if ( x < 0 ) x = 0;
		if ( y < 0 ) y = 0;
		if ( x+w >= m_options.size.width ) w = m_options.size.width-x;
		if ( y+h >= m_options.size.height ) h = m_options.size.height-y;

		for ( int i=x; i<x+w; i++ )
		{
			for ( int j=y; j<y+h; j++ )
			{
				Cell& c = m_scene.backbuffer[m_options.size.width*y+x];
				c.tiles.clear();
				c.background = m_scene.background;
			}
		}
	}

	void Terminal::SetColor(Color color)
	{
		m_scene.foreground = color;
	}

	void Terminal::SetBkColor(Color color)
	{
		m_scene.background = color;
	}

	void Terminal::SetBlending(int mode)
	{
		m_scene.blending = mode;
	}

	void Terminal::PutUnlocked(int x, int y, int code)
	{
		if ( x < 0 || x >= m_options.size.width || y < 0 || y >= m_options.size.height )
		{
			return;
		}

		Cell& c = m_scene.backbuffer[m_options.size.width*y+x];
		code = m_encoding.ToUCS2(code);

		if ( code == 0 )
		{
			c.tiles.clear();
		}
		else
		{
			if (m_scene.background.a > 0)
			{
				c.background = m_scene.background;
			}

			if ( m_scene.blending == BLENDING_NONE && !c.tiles.empty() )
			{
				c.tiles.clear();
			}

			c.tiles.push_back(Cell::Tile(code, m_scene.foreground));
		}
	}

	void Terminal::Put(int x, int y, int code)
	{
		std::lock_guard<std::mutex> guard(m_lock); // Call modifies the scene
		PutUnlocked(x, y, code);
	}

	int Terminal::Print(int x0, int y0, const char* s8)
	{
		std::lock_guard<std::mutex> guard(m_lock); // Call modifies the scene

		std::wstring s = m_encoding.ToUCS2(s8);

		int x = x0, y = y0;
		int printed = 0;
		
		color_t foreground = m_scene.foreground;
		color_t background = m_scene.background;

		Size size = m_options.size;

		const auto put_and_increment = [&](wchar_t code)
		{
			PutUnlocked(x, y, code);
			printed += 1;

			if ( ++x >= size.width )
			{
				x = 0;
				if ( ++y >= size.height )
				{
					y = 0;
				}
			}
		};

		const auto apply_tag = [&](const std::wstring& s, size_t begin, size_t end)
		{
			if ( s[begin+1] == L'/' )
			{
				// Cancel tag: [/name]

				std::wstring name = s.substr(begin+2, end-(begin+2));

				if ( name == L"color" )
				{
					SetColor(foreground);
				}
				else if ( name == L"bkcolor" )
				{
					SetBkColor(background);
				}
			}
			else
			{
				// Set tag: [name] or [name=value]

				std::wstring name, value;

				size_t n_equals = s.find(L'=', begin);
				if ( n_equals == std::wstring::npos || n_equals > end-1 )
				{
					name = s.substr(begin+1, end-(begin+1));
				}
				else
				{
					name = s.substr(begin+1, n_equals-(begin+1));
					value = s.substr(n_equals+1, end-(n_equals+1));
				}

				if ( name.length() == 0 ) return;

				if ( name == L"color" )
				{
					SetColor(Palette::Instance[value]);
				}
				else if ( name == L"bkcolor" )
				{
					SetBkColor(Palette::Instance[value]);
				}
				else if ( name[0] == 'u' || name[0] == 'U' )
				{
					if ( name.length() > 1 )
					{
						std::wstringstream stream;
						stream << std::hex;
						stream << name.substr(1);
						uint16_t value = 0;
						stream >> value;
						put_and_increment(value);
					}
				}
			}
		};

		for ( size_t i = 0; i < s.length(); i++ )
		{
			wchar_t c = s[i];

			if ( c == L'[' && m_options.output_postformat )
			{
				if ( i == s.length()-1 ) break; // Malformed postformatting tag
				
				if ( s[i+1] == L'[' )
				{
					// Escaped '['
					i += 1;
					put_and_increment(s[i]);
				}
				else
				{
					// Start of a postformatting tag
					size_t closing = s.find(L']', i);
					if ( closing == std::wstring::npos ) break; // Malformed, no closing ']'
					apply_tag(s, i, closing);
					i = closing;
				}
			}
			else if ( c == L']' && m_options.output_postformat )
			{
				// This MUST be an escaped ']' because regular closing postformatting ']' will be
				// consumed while parsing that tag

				if ( i == s.length()-1 ) break; // Malformed

				i += 1;
				put_and_increment(s[i]);
			}
			else if ( c == '\n' )
			{
				x = x0;
				y += 1;
			}
			else
			{
				put_and_increment(c);
			}
		}

		// Revert state
		m_scene.foreground = foreground;
		m_scene.background = background;

		return printed;
	}

	int Terminal::HasInput()
	{
		if ( m_status == kUninitialized ) return 0;
		if ( m_status == kClosed ) return 1;

		std::lock_guard<std::mutex> guard(m_lock);

		if ( m_state[VK_CLOSE] && m_options.input_sticky_close ) return 1; // sticky VK_CLOSE, once set, can't be undone
		if ( m_input_queue.empty() ) return 0;
		ConsumeIrrelevantInput(); // FIXME: unnecessary?

		return !m_input_queue.empty();
	}

	int Terminal::GetState(int code)
	{
		std::lock_guard<std::mutex> guard(m_lock);
		if ( code < 0 || code > (int)m_state.size() ) return 0;
		return m_state[code];
	}

	int Terminal::Get()
	{
		Keystroke stroke = GetStroke();
		int result = stroke.scancode;

		if ( result != VK_MOUSE_MOVE && result != VK_MOUSE_SCROLL && !stroke.pressed )
		{
			// Not movement but a released key
			result |= VK_FLAG_RELEASED;
		}

		return result;
	}

	int Terminal::GetStr(int x, int y, char* buffer, int max)
	{
		return 0; // NYI, heavy wrapper around Get
	}

	void Terminal::Render() // NOTE: This method is ALWAYS called from rendering thread and with mutex already locked
	{
		if ( !m_scene.font )
		{
			LOG(Error, "No FontProvider has been object allocated before render");
			return;
		}

		if ( !m_scene.initialized )
		{
			ConfigureViewport();
			m_scene.initialized = true;
		}

		//if ( m_new_characters ) // TODO: optimization: new characters flag
		{
			std::list<std::uint16_t> need_preparation;
			for ( Cell& cell: m_scene.frontbuffer )
			{
				for ( Cell::Tile& tile: cell.tiles )
				{
					wchar_t code = tile.code;

					if ( code == 0 ) continue;
					
					if ( !m_scene.tileset.HasTileFor(code) )
					{
						if ( m_scene.font->HasGlyph(code) )
						{
							need_preparation.push_back(code);
						}
						else
						{
							m_scene.tileset.Add(code, 0xFFFF);
						}
					}
				}
			}

			if ( !need_preparation.empty() )
			{
				m_scene.font->PopulateTileset(m_scene.tileset, need_preparation);
			}
		}

		// Draw
		glClear(GL_COLOR_BUFFER_BIT);
		glLoadIdentity();

		RenderBackgrounds();
		RenderGlyphs();

		/*
		// DEBUG
		glColor4f(1, 1, 1, 1);
		glBegin(GL_QUADS);
		glTexCoord2f(0, 0); glVertex2i(0, 0);
		glTexCoord2f(0, 1); glVertex2i(0, 256);
		glTexCoord2f(1, 1); glVertex2i(512, 256);
		glTexCoord2f(1, 0); glVertex2i(512, 0);
		glEnd();
		//*/
	}

	void Terminal::ConfigureViewport()
	{
		Size size = m_options.size*m_scene.font->GetCellSize();

		glClearColor(0, 0, 0, 1);

		glViewport(0, 0, size.width, size.height);

		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glOrtho(0, size.width, size.height, 0, -1, +1);

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();

		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	}

	void Terminal::RenderBackgrounds()
	{
		Size size = m_options.size;
		Size cell_size = m_scene.font->GetCellSize();

		// Background colors
		Texture::Disable();
		glBegin(GL_QUADS);
		for ( int x = 0; x < size.width; x++ )
		{
			for ( int y = 0; y < size.height; y++ )
			{
				Cell& cell = m_scene.frontbuffer[y*size.width+x];

				Color bkcolor = cell.background;
				if ( bkcolor.r || bkcolor.g || bkcolor.b )
				{
					glColor3ub(bkcolor.r, bkcolor.g, bkcolor.b );
					int x0 = x * cell_size.width, y0 = y * cell_size.height;
					int x1 = x0 + cell_size.width, y1 = y0 + cell_size.height;
					glVertex2i(x0, y0);
					glVertex2i(x0, y1);
					glVertex2i(x1, y1);
					glVertex2i(x1, y0);
				}
			}
		}
		glEnd();
	}

	void Terminal::RenderGlyphs()
	{
		Texture::Enable();
		m_scene.tileset.Bind();

		Size size = m_options.size;
		Size cell_size = m_scene.font->GetCellSize();

		glBegin(GL_QUADS);
		for ( int x = 0; x < size.width; x++ )
		{
			for ( int y = 0; y < size.height; y++ )
			{
				Cell& cell = m_scene.frontbuffer[y*size.width+x];

				for ( Cell::Tile& tile: cell.tiles )
				{
					Color color = tile.foreground;
					glColor4ub(color.r, color.g, color.b, color.a);

					uint16_t code = tile.code; // TODO: wchar_t -> uint16_t
					if (m_scene.sprites.HasSprite(code))
					{
						// Inject image glyph into regular glyph rendering
						glEnd();
						m_scene.sprites.Draw(code, Point(x, y), cell_size);
						m_scene.tileset.Bind();
						glBegin(GL_QUADS);
					}
					else
					{
						m_scene.tileset.Draw(code, Point(x, y)*cell_size);
					}
				}
			}
		}
		glEnd();
	}

	void Terminal::OnWindowInput(Keystroke keystroke) // FIXME: const&
	{
		std::lock_guard<std::mutex> guard(m_lock);

		if ( m_options.input_events != InputFilter::None )
		{
			m_input_queue.push_back(keystroke);
			ConsumeIrrelevantInput();
			if ( !m_input_queue.empty() )
			{
				m_input_condvar.notify_all();
			}
		}
		else
		{
			ConsumeStroke(keystroke);
		}
	}

	void Terminal::OnWindowActivate()
	{
		for ( int i = 0; i < VK_F12; i++ ) // FIXME: VK_F12 --> some meaningful name
		{
			if ( i == VK_CLOSE ) continue;
			if ( GetState(i) > 0 ) OnWindowInput(Keystroke(i, false));
		}
	}

	void Terminal::OnWindowRedraw()
	{
		// Ignore irrelevant redraw calls
		if ( m_status != kNormal ) return;
		//if ( m_flag_enforced_redraw ) return;

		custom_render_cb_t custom_render = nullptr;

		// Lock with narrowed scope
		{
			/*
			std::unique_lock<std::mutex> guard(m_lock, std::try_to_lock);
			if ( !guard.owns_lock() ) return;
			*/
			std::unique_lock<std::mutex> guard(m_lock);

			// In synchronous mode only first redraw is allowed
			//if ( m_options.output_synchronous && !m_flag_enforced_redraw ) return;

			// Draw scene under the already aquired lock
			Render();

			custom_render = m_options.output_custom_render;
		}

		if ( custom_render != nullptr )
		{
			ScopedSet<volatile bool> guard(m_flag_custom_render, true, false);
			custom_render();
		}
	}

	void Terminal::OnWindowClose()
	{
		LOG(Trace, "OnWindowClose callback is called");

		// NOTE: if window initialization has succeeded, this callback will be called for sure
		m_status = kClosed;

		std::lock_guard<std::mutex> guard(m_lock);

		// Dispose of graphics
		m_scene.tileset.Dispose();
		m_scene.sprites.Clear();
		Texture::Unbind();

		// Unblock possibly blocked client thread
		m_input_condvar.notify_all();
	}

	void Terminal::ConsumeStroke(const Keystroke& stroke)
	{
		if ( stroke.scancode == VK_MOUSE_MOVE )
		{
			// Mouse movement event, update mouse position
			float cell_width = m_state[VK_CELL_WIDTH];
			float cell_height = m_state[VK_CELL_HEIGHT];
			m_state[VK_MOUSE_X] = (int)std::floor(stroke.x/cell_width);
			m_state[VK_MOUSE_Y] = (int)std::floor(stroke.y/cell_height);
			m_state[VK_MOUSE_PIXEL_X] = stroke.x;
			m_state[VK_MOUSE_PIXEL_Y] = stroke.y;
		}
		else if ( stroke.scancode == VK_MOUSE_SCROLL )
		{
			// Mouse scroll event, update wheel position
			m_state[VK_MOUSE_WHEEL] = stroke.z; // FIXME: relative
		}
		else if ( stroke.scancode == VK_CLOSE )
		{
			if ( m_options.input_sticky_close || !m_options.input_events.Has(InputFilter::KeyPress) )
			{
				m_state[VK_CLOSE] = 1;
			}
		}
		else
		{
			m_state[stroke.scancode] = (int)stroke.pressed;
		}
	}

	void Terminal::ConsumeIrrelevantInput()
	{
		//InputEvents::BitmaskType filter = m_options.input_events;
		InputFilter& filter = m_options.input_events;

		while ( !m_input_queue.empty() )
		{
			Keystroke& stroke = m_input_queue.front();
			bool must_be_consumed = false;

			if ( stroke.scancode == VK_MOUSE_MOVE )
			{
				if ( !filter.Has(InputFilter::MouseMove) ) must_be_consumed = true;
			}
			else if ( stroke.scancode == VK_MOUSE_SCROLL )
			{
				if ( !filter.Has(InputFilter::MouseScroll) ) must_be_consumed = true;
			}
			else
			{
				if ( stroke.pressed )
				{
					if ( !filter.Has(InputFilter::KeyPress) ) must_be_consumed = true;
				}
				else
				{
					if ( !filter.Has(InputFilter::KeyRelease) ) must_be_consumed = true;
				}
			}

			if ( !must_be_consumed ) break;

			ConsumeStroke(stroke);
			m_input_queue.pop_front();
			continue;
		}
	}

	Keystroke Terminal::GetStroke()
	{
		if ( m_status == kUninitialized ) return Keystroke(0); // Not-a-keystroke
		if ( m_status == kClosed ) return Keystroke(VK_CLOSE);

		std::unique_lock<std::mutex> lock(m_lock);

		if ( m_state[VK_CLOSE] && m_options.input_sticky_close ) return VK_CLOSE; // sticky VK_CLOSE, once set, can't be undone

		if ( m_options.input_events == InputFilter::None )
		{
			// Keep terminal_get from accidental eternal block while input.events == none
			return 0;
		}

		ConsumeIrrelevantInput();

		const auto retrieve_keystroke = [&]() -> Keystroke
		{
			Keystroke result = m_input_queue.front();
			m_input_queue.pop_front();
			ConsumeStroke(result);
			return result;
		};

		if ( m_options.input_timeout == Timeout::Immediate ) // Return immediately
		{
			return m_input_queue.empty()? Keystroke(0): retrieve_keystroke();
		}
		
		auto predicate = [&]() -> bool { return !m_input_queue.empty() || (m_status != kNormal); };

		int32_t interval =
			m_options.input_timeout == Timeout::Indefinite?
			std::numeric_limits<int32_t>::max():
			(int)m_options.input_timeout;

		std::chrono::milliseconds timeout(interval);

		if ( !m_input_condvar.wait_for(lock, timeout, predicate) )
		{
			// Timeout expired
			return Keystroke(0); // Not-a-keystroke FIXME: expired!
		}
		else
		{
			return retrieve_keystroke();
		}
	}

	int Terminal::UseImage(int code, const char* filename)
	{
		// NOTE: Image glyph manipulation modifies textures and must be executed on window thread

		if (filename != nullptr)
		{
			std::wstring filename_u16 = m_encoding.ToUCS2(filename);
			bool result = false;
			m_window->Invoke([&](){result = m_scene.sprites.Add(code, filename_u16);});
			return result? 0: -1;
		}
		else
		{
			m_window->Invoke([&](){m_scene.sprites.Remove(code);});
			return 0;
		}
	}
}

