/*
* BearLibTerminal
* Copyright (C) 2013 Cfyz
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
* FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
* COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
* IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
* CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include <stdexcept>
#include <cstdlib>
#include "Tileset.hpp"
#include "Utility.hpp"
#include "Geometry.hpp"
#include "OpenGL.hpp"
#include "Log.hpp"

namespace BearLibTerminal
{
	Tileset::Tileset()
	{ }

	Size Tileset::GetTileSize() const
	{
		return m_tile_size;
	}

	size_t Tileset::GetTileCount() const
	{
		return m_coords.size();
	}

	void Tileset::Clear()
	{
		m_canvas.Dispose();
		m_tile_size = Size();
		m_coords.clear();
		m_free_slots.clear();
		m_canvas_is_dirty = false;
	}

	void Tileset::Dispose()
	{
		Clear();
		m_texture.Dispose();
	}

	size_t Tileset::Add(const Bitmap& tile)
	{
		return Add(tile, Rectangle(tile.GetSize()));
	}

	size_t Tileset::Add(const Bitmap& tile, const Rectangle& area)
	{
		if ( m_tile_size.Area() == 0 )
		{
			m_tile_size = area.Size();
		}
		else if ( area.Size() != m_tile_size )
		{
			throw std::runtime_error("invalid tile size");
		}

		Point location;

		if ( m_free_slots.empty() )
		{
			auto tile_capacity = [=](Size canvas_size) -> Size
			{
				return Size
					(
						(size_t)std::floor(canvas_size.width/m_tile_size.width),
						(size_t)std::floor(canvas_size.height/m_tile_size.height)
					);
			};

			if ( m_canvas.IsEmpty() )
			{
				// Allocate initial
				m_canvas = Bitmap(Size(256, 256));
				Size capacity = tile_capacity(m_canvas.GetSize());
				for_each_point(Rectangle(capacity), [&](Point pt) -> void { m_free_slots.push_back(pt); });
			}
			else
			{
				// Expand
				Size new_size = m_canvas.GetSize();
				(new_size.width == new_size.height)? new_size.width *= 2: new_size.height *= 2;

				Bitmap new_canvas(new_size);
				new_canvas.Blit(m_canvas, Point(0, 0));
				Size old_capacity = tile_capacity(m_canvas.GetSize());
				Size new_capacity = tile_capacity(new_canvas.GetSize());

				Rectangle free_area =
					(new_capacity.width > old_capacity.width)?
					Rectangle(old_capacity.width, 0, (new_capacity.width-old_capacity.width), new_capacity.height):
					Rectangle(0, old_capacity.height, new_capacity.width, (new_capacity.height-old_capacity.height));

				for_each_point(free_area, [&](Point pt) { m_free_slots.push_back(pt); });

				m_canvas = new_canvas;
			}

			m_relative_tile_size = As<float>(m_tile_size) / As<float>(m_canvas.GetSize());
		}
		
		location = m_free_slots.front();
		m_free_slots.pop_front();

		Rectangle slot_coords(location*m_tile_size, m_tile_size);

		m_canvas.Blit(tile, area, slot_coords.Location());
		m_coords.push_back(location);
		m_canvas_is_dirty = true;

		return (m_coords.size()-1);
	}

	void Tileset::Bind()
	{
		if ( m_canvas_is_dirty )
		{
			m_texture.Update(m_canvas);
			m_canvas_is_dirty = false;
		}

		m_texture.Bind();
	}

	void Tileset::Draw(size_t index, Point location) const
	{
		if ( index >= m_coords.size() )
		{
			throw std::runtime_error("index out of bounds");
		}

		const Point& slot = m_coords[index];

		float tu0 = slot.x * m_relative_tile_size.width;
		float tv0 = slot.y * m_relative_tile_size.height;
		float tu1 = tu0 + m_relative_tile_size.width;
		float tv1 = tv0 + m_relative_tile_size.height;

		glTexCoord2f(tu0, tv0);
		glVertex2i(location.x, location.y);

		glTexCoord2f(tu0, tv1);
		glVertex2i(location.x, location.y+m_tile_size.height);

		glTexCoord2f(tu1, tv1);
		glVertex2i(location.x+m_tile_size.width, location.y+m_tile_size.height);
		
		glTexCoord2f(tu1, tv0);
		glVertex2i(location.x+m_tile_size.width, location.y);
	}
}
