/*
* BearLibTerminal
* Copyright (C) 2013 Cfyz
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
* FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
* COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
* IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
* CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifdef __linux__

#include "X11Window3.hpp"
#include "OpenGL.hpp"
#include "Log.hpp"
#include "Encoding.hpp"
#include <X11/Xlib.h>
#include <X11/Xmu/Xmu.h>
#include <GL/glx.h>
#include <unistd.h>

namespace BearLibTerminal
{
	struct X11Window::Private
	{
		Private();

		Display* display;
		::Window window;
		int screen;
		XVisualInfo* visual;
		GLXContext glx;
		XIM im;
		XIC ic;
	};

	X11Window::Private::Private():
		display(NULL),
		window(0),
		screen(0),
		visual(NULL),
		glx(NULL)
	{ }

	// attributes for a single buffered visual in RGBA format with at least
	// 4 bits per color and a 16 bit depth buffer
	static int glx_single_buffered_attrs[] =
	{
		GLX_RGBA,
		GLX_RED_SIZE, 4,
		GLX_GREEN_SIZE, 4,
		GLX_BLUE_SIZE, 4,
		GLX_DEPTH_SIZE, 16,
		None
	};

	// attributes for a double buffered visual in RGBA format with at least
	// 4 bits per color and a 16 bit depth buffer
	static int glx_double_buffered_attrs[] =
	{
		GLX_RGBA,
		GLX_DOUBLEBUFFER,
		GLX_RED_SIZE, 4,
		GLX_GREEN_SIZE, 4,
		GLX_BLUE_SIZE, 4,
		GLX_DEPTH_SIZE, 16,
		None
	};

	static XIMStyle ChooseBetterStyle(XIMStyle style1, XIMStyle style2)
	{
		XIMStyle s,t;
		XIMStyle preedit = XIMPreeditArea | XIMPreeditCallbacks | XIMPreeditPosition | XIMPreeditNothing | XIMPreeditNone;
		XIMStyle status = XIMStatusArea | XIMStatusCallbacks | XIMStatusNothing | XIMStatusNone;
		if (style1 == 0) return style2;
		if (style2 == 0) return style1;
		if ((style1 & (preedit | status)) == (style2 & (preedit | status))) return style1;
		s = style1 & preedit;
		t = style2 & preedit;
		if (s != t)
		{
			if (s | t | XIMPreeditCallbacks)
				return (s == XIMPreeditCallbacks)?style1:style2;
			else if (s | t | XIMPreeditPosition)
				return (s == XIMPreeditPosition)?style1:style2;
			else if (s | t | XIMPreeditArea)
				return (s == XIMPreeditArea)?style1:style2;
			else if (s | t | XIMPreeditNothing)
				return (s == XIMPreeditNothing)?style1:style2;
		}
		else
		{
			// if preedit flags are the same, compare status flags
			s = style1 & status;
			t = style2 & status;
			if (s | t | XIMStatusCallbacks)
				return (s == XIMStatusCallbacks)?style1:style2;
			else if (s | t | XIMStatusArea)
				return (s == XIMStatusArea)?style1:style2;
			else if (s | t | XIMStatusNothing)
				return (s == XIMStatusNothing)?style1:style2;
		}

		return 0;
	}

	X11Window::X11Window():
		m_private(new Private())
	{ }

	X11Window::~X11Window()
	{
		LOG(Trace, "~X11Window");
		Stop();
		LOG(Trace, "Done with window object (X11)");
	}

	bool X11Window::ValidateIcon(const std::wstring& filename)
	{
		// TODO: Learn about working with icons in Linux
		return true;
	}

	void X11Window::SetTitle(const std::wstring& title)
	{
		std::lock_guard<std::mutex> guard(m_lock);
		if ( m_private->window == 0 ) return;
		std::string u8 = Encoding::UTF8.FromUCS2(title);
		XChangeProperty
		(
			m_private->display,
			m_private->window,
			XInternAtom(m_private->display, "_NET_WM_NAME", false),
			XInternAtom(m_private->display, "UTF8_STRING",  false),
			8,
			PropModeReplace,
			(const unsigned char*)u8.c_str(),
			u8.size()
		);
	}

	void X11Window::SetIcon(const std::wstring& filename)
	{
		// TODO: Learn how to set icons to apps in Linux
	}

	void X11Window::SetClientSize(const Size& size)
	{
		std::lock_guard<std::mutex> guard(m_lock);
		if ( m_private->window == 0 ) return;
		XResizeWindow(m_private->display, m_private->window, size.width, size.height);
	}

	void X11Window::ForceRedraw()
	{
		std::lock_guard<std::mutex> guard(m_lock);
		XClearArea(m_private->display, m_private->window, 0, 0, 640, 480, True);
		m_redraw_barrier.Wait();
	}

	void X11Window::Show()
	{
		std::lock_guard<std::mutex> guard(m_lock);
		if ( m_private->window != 0 ) XMapWindow(m_private->display, m_private->window);
	}

	void X11Window::Hide()
	{
		std::lock_guard<std::mutex> guard(m_lock);
		if ( m_private->window != 0 ) XUnmapWindow(m_private->display, m_private->window);
	}

	void X11Window::HandleRepaint()
	{
		try
		{
			InvokeCallback(m_on_redraw);
			glXSwapBuffers(m_private->display, m_private->window);
		}
		catch ( std::exception& e )
		{
			LOG(Fatal, L"Rendering routine has thrown an exception: " << e.what());
			m_proceed = false;
		}
		catch ( ... )
		{
			LOG(Fatal, L"Rendering routine has thrown an exception of an unknown type");
			m_proceed = false;
		}

		// Open barrier
		m_redraw_barrier.Notify();
	}

	/*
	static SDLKey ODD_keymap[256];
	static SDLKey MISC_keymap[256];
	SDLKey X11_TranslateKeycode(Display *display, KeyCode kc);
	*/

	/*
	void X11_InitKeymap(void)
	{
		int i;

		// Odd keys used in international keyboards
		for ( i=0; i<SDL_arraysize(ODD_keymap); ++i )
			ODD_keymap[i] = SDLK_UNKNOWN;

		// Some of these might be mappable to an existing SDLK_ code
		ODD_keymap[XK_dead_grave&0xFF] = SDLK_COMPOSE;
		ODD_keymap[XK_dead_acute&0xFF] = SDLK_COMPOSE;
		ODD_keymap[XK_dead_tilde&0xFF] = SDLK_COMPOSE;
		ODD_keymap[XK_dead_macron&0xFF] = SDLK_COMPOSE;
		ODD_keymap[XK_dead_breve&0xFF] = SDLK_COMPOSE;
		ODD_keymap[XK_dead_abovedot&0xFF] = SDLK_COMPOSE;
		ODD_keymap[XK_dead_diaeresis&0xFF] = SDLK_COMPOSE;
		ODD_keymap[XK_dead_abovering&0xFF] = SDLK_COMPOSE;
		ODD_keymap[XK_dead_doubleacute&0xFF] = SDLK_COMPOSE;
		ODD_keymap[XK_dead_caron&0xFF] = SDLK_COMPOSE;
		ODD_keymap[XK_dead_cedilla&0xFF] = SDLK_COMPOSE;
		ODD_keymap[XK_dead_ogonek&0xFF] = SDLK_COMPOSE;
		ODD_keymap[XK_dead_iota&0xFF] = SDLK_COMPOSE;
		ODD_keymap[XK_dead_voiced_sound&0xFF] = SDLK_COMPOSE;
		ODD_keymap[XK_dead_semivoiced_sound&0xFF] = SDLK_COMPOSE;
		ODD_keymap[XK_dead_belowdot&0xFF] = SDLK_COMPOSE;
	#ifdef XK_dead_hook
		ODD_keymap[XK_dead_hook&0xFF] = SDLK_COMPOSE;
	#endif
	#ifdef XK_dead_horn
		ODD_keymap[XK_dead_horn&0xFF] = SDLK_COMPOSE;
	#endif

	#ifdef XK_dead_circumflex
		// These X keysyms have 0xFE as the high byte
		ODD_keymap[XK_dead_circumflex&0xFF] = SDLK_CARET;
	#endif
	#ifdef XK_ISO_Level3_Shift
		ODD_keymap[XK_ISO_Level3_Shift&0xFF] = SDLK_MODE; // "Alt Gr" key
	#endif

		// Map the miscellaneous keys
		for ( i=0; i<SDL_arraysize(MISC_keymap); ++i )
			MISC_keymap[i] = SDLK_UNKNOWN;

		// These X keysyms have 0xFF as the high byte
		MISC_keymap[XK_BackSpace&0xFF] = SDLK_BACKSPACE;
		MISC_keymap[XK_Tab&0xFF] = SDLK_TAB;
		MISC_keymap[XK_Clear&0xFF] = SDLK_CLEAR;
		MISC_keymap[XK_Return&0xFF] = SDLK_RETURN;
		MISC_keymap[XK_Pause&0xFF] = SDLK_PAUSE;
		MISC_keymap[XK_Escape&0xFF] = SDLK_ESCAPE;
		MISC_keymap[XK_Delete&0xFF] = SDLK_DELETE;

		MISC_keymap[XK_KP_0&0xFF] = SDLK_KP0;		// Keypad 0-9
		MISC_keymap[XK_KP_1&0xFF] = SDLK_KP1;
		MISC_keymap[XK_KP_2&0xFF] = SDLK_KP2;
		MISC_keymap[XK_KP_3&0xFF] = SDLK_KP3;
		MISC_keymap[XK_KP_4&0xFF] = SDLK_KP4;
		MISC_keymap[XK_KP_5&0xFF] = SDLK_KP5;
		MISC_keymap[XK_KP_6&0xFF] = SDLK_KP6;
		MISC_keymap[XK_KP_7&0xFF] = SDLK_KP7;
		MISC_keymap[XK_KP_8&0xFF] = SDLK_KP8;
		MISC_keymap[XK_KP_9&0xFF] = SDLK_KP9;
		MISC_keymap[XK_KP_Insert&0xFF] = SDLK_KP0;
		MISC_keymap[XK_KP_End&0xFF] = SDLK_KP1;
		MISC_keymap[XK_KP_Down&0xFF] = SDLK_KP2;
		MISC_keymap[XK_KP_Page_Down&0xFF] = SDLK_KP3;
		MISC_keymap[XK_KP_Left&0xFF] = SDLK_KP4;
		MISC_keymap[XK_KP_Begin&0xFF] = SDLK_KP5;
		MISC_keymap[XK_KP_Right&0xFF] = SDLK_KP6;
		MISC_keymap[XK_KP_Home&0xFF] = SDLK_KP7;
		MISC_keymap[XK_KP_Up&0xFF] = SDLK_KP8;
		MISC_keymap[XK_KP_Page_Up&0xFF] = SDLK_KP9;
		MISC_keymap[XK_KP_Delete&0xFF] = SDLK_KP_PERIOD;
		MISC_keymap[XK_KP_Decimal&0xFF] = SDLK_KP_PERIOD;
		MISC_keymap[XK_KP_Divide&0xFF] = SDLK_KP_DIVIDE;
		MISC_keymap[XK_KP_Multiply&0xFF] = SDLK_KP_MULTIPLY;
		MISC_keymap[XK_KP_Subtract&0xFF] = SDLK_KP_MINUS;
		MISC_keymap[XK_KP_Add&0xFF] = SDLK_KP_PLUS;
		MISC_keymap[XK_KP_Enter&0xFF] = SDLK_KP_ENTER;
		MISC_keymap[XK_KP_Equal&0xFF] = SDLK_KP_EQUALS;

		MISC_keymap[XK_Up&0xFF] = SDLK_UP;
		MISC_keymap[XK_Down&0xFF] = SDLK_DOWN;
		MISC_keymap[XK_Right&0xFF] = SDLK_RIGHT;
		MISC_keymap[XK_Left&0xFF] = SDLK_LEFT;
		MISC_keymap[XK_Insert&0xFF] = SDLK_INSERT;
		MISC_keymap[XK_Home&0xFF] = SDLK_HOME;
		MISC_keymap[XK_End&0xFF] = SDLK_END;
		MISC_keymap[XK_Page_Up&0xFF] = SDLK_PAGEUP;
		MISC_keymap[XK_Page_Down&0xFF] = SDLK_PAGEDOWN;

		MISC_keymap[XK_F1&0xFF] = SDLK_F1;
		MISC_keymap[XK_F2&0xFF] = SDLK_F2;
		MISC_keymap[XK_F3&0xFF] = SDLK_F3;
		MISC_keymap[XK_F4&0xFF] = SDLK_F4;
		MISC_keymap[XK_F5&0xFF] = SDLK_F5;
		MISC_keymap[XK_F6&0xFF] = SDLK_F6;
		MISC_keymap[XK_F7&0xFF] = SDLK_F7;
		MISC_keymap[XK_F8&0xFF] = SDLK_F8;
		MISC_keymap[XK_F9&0xFF] = SDLK_F9;
		MISC_keymap[XK_F10&0xFF] = SDLK_F10;
		MISC_keymap[XK_F11&0xFF] = SDLK_F11;
		MISC_keymap[XK_F12&0xFF] = SDLK_F12;
		MISC_keymap[XK_F13&0xFF] = SDLK_F13;
		MISC_keymap[XK_F14&0xFF] = SDLK_F14;
		MISC_keymap[XK_F15&0xFF] = SDLK_F15;

		MISC_keymap[XK_Num_Lock&0xFF] = SDLK_NUMLOCK;
		MISC_keymap[XK_Caps_Lock&0xFF] = SDLK_CAPSLOCK;
		MISC_keymap[XK_Scroll_Lock&0xFF] = SDLK_SCROLLOCK;
		MISC_keymap[XK_Shift_R&0xFF] = SDLK_RSHIFT;
		MISC_keymap[XK_Shift_L&0xFF] = SDLK_LSHIFT;
		MISC_keymap[XK_Control_R&0xFF] = SDLK_RCTRL;
		MISC_keymap[XK_Control_L&0xFF] = SDLK_LCTRL;
		MISC_keymap[XK_Alt_R&0xFF] = SDLK_RALT;
		MISC_keymap[XK_Alt_L&0xFF] = SDLK_LALT;
		MISC_keymap[XK_Meta_R&0xFF] = SDLK_RMETA;
		MISC_keymap[XK_Meta_L&0xFF] = SDLK_LMETA;
		MISC_keymap[XK_Super_L&0xFF] = SDLK_LSUPER; // Left "Windows"
		MISC_keymap[XK_Super_R&0xFF] = SDLK_RSUPER; // Right "Windows
		MISC_keymap[XK_Mode_switch&0xFF] = SDLK_MODE; // "Alt Gr" key
		MISC_keymap[XK_Multi_key&0xFF] = SDLK_COMPOSE; // Multi-key compose

		MISC_keymap[XK_Help&0xFF] = SDLK_HELP;
		MISC_keymap[XK_Print&0xFF] = SDLK_PRINT;
		MISC_keymap[XK_Sys_Req&0xFF] = SDLK_SYSREQ;
		MISC_keymap[XK_Break&0xFF] = SDLK_BREAK;
		MISC_keymap[XK_Menu&0xFF] = SDLK_MENU;
		MISC_keymap[XK_Hyper_R&0xFF] = SDLK_MENU; // Windows "Menu" key
	}
	*/

	/*
	SDLKey X11_TranslateKeycode(Display *display, KeyCode kc)
	{
		KeySym xsym;
		SDLKey key;

		xsym = XKeycodeToKeysym(display, kc, 0);
	#ifdef DEBUG_KEYS
		fprintf(stderr, "Translating key code %d -> 0x%.4x\n", kc, xsym);
	#endif
		key = SDLK_UNKNOWN;
		if ( xsym ) {
			switch (xsym>>8) {
				case 0x1005FF:
	#ifdef SunXK_F36
				if ( xsym == SunXK_F36 )
					key = SDLK_F11;
	#endif
	#ifdef SunXK_F37
				if ( xsym == SunXK_F37 )
					key = SDLK_F12;
	#endif
				break;
				case 0x00:	// Latin 1
				key = (SDLKey)(xsym & 0xFF);
				break;
				case 0x01:	// Latin 2
				case 0x02:	// Latin 3
				case 0x03:	// Latin 4
				case 0x04:	// Katakana
				case 0x05:	// Arabic
				case 0x06:	// Cyrillic
				case 0x07:	// Greek
				case 0x08:	// Technical
				case 0x0A:	// Publishing
				case 0x0C:	// Hebrew
				case 0x0D:	// Thai
				// These are wrong, but it's better than nothing
				key = (SDLKey)(xsym & 0xFF);
				break;
				case 0xFE:
				key = ODD_keymap[xsym&0xFF];
				break;
				case 0xFF:
				key = MISC_keymap[xsym&0xFF];
				break;
				default:
				//
				//fprintf(stderr, "X11: Unhandled xsym, sym = 0x%04x\n",
				//		(unsigned int)xsym);
				//
				break;
			}
		} else {
			// X11 doesn't know how to translate the key!
			switch (kc) {
				// Caution:
				//   These keycodes are from the Microsoft Keyboard
				//
				case 115:
				key = SDLK_LSUPER;
				break;
				case 116:
				key = SDLK_RSUPER;
				break;
				case 117:
				key = SDLK_MENU;
				break;
				default:
				//
				// no point in an error message; happens for
				// several keys when we get a keymap notify
				//
				break;
			}
		}
		return key;
	}
	*/

	void X11Window::HandleKey()
	{

	}

	struct InvokationSentry
	{
		std::function<void()> func;
		Semaphore semaphore;
	};

	void X11Window::Invoke(std::function<void()> func)
	{
		InvokationSentry sentry;
		sentry.func = func;

		XClientMessageEvent event;
		memset(&event, 0, sizeof(XClientMessageEvent));
		event.type = ClientMessage;
		event.window = m_private->window;
		event.format = 32;
		// FIXME: this is strictly 32-bit!
		event.data.l[0] = (long)this; // Mark
		event.data.l[4] = (long)&sentry;
		XSendEvent(m_private->display, m_private->window, 0, 0, (XEvent*)&event);
		XFlush(m_private->display);

		sentry.semaphore.Wait();
	}

	void X11Window::ThreadFunction()
	{
		LOG(Trace, "Entering X11-specific window thread function");

		XEvent e;

		Atom wmDeleteMessage = XInternAtom(m_private->display, "WM_DELETE_WINDOW", False);
		XSetWMProtocols(m_private->display, m_private->window, &wmDeleteMessage, 1);

		while ( m_proceed )
		{
			if ( XPending(m_private->display) )
			{
				XNextEvent(m_private->display, &e);

				if ( (e.type == Expose && e.xexpose.count == 0) || (e.type == KeyPress) )
				{
					// draw or redraw the window
					HandleRepaint();
				}
				else if ( e.type == KeyPress )
				{
					wchar_t buffer[255];
					KeySym key;
					Status status;

					int rc = XwcLookupString(m_private->ic, &e.xkey, buffer, 255, &key, &status);
					if ( rc > 0 )
					{
						// Translated unicode, has vk and character code
						//
						//for ( int i=0; i<rc; i++ )
						//{
						//	printf("#%04X ", (unsigned int)buffer[i]);
						//}
						//printf("[s=%d]", e.xkey.state); //LockMask;
						//printf("\n");
					}
					else
					{
						// Functional keys such as Ctrl, Alt, F1, etc., has only vk
						//
						//printf("#%04X (ASCII) ", e.xkey.keycode);
						//printf("\n");
					}
				}
				else if ( e.type == ConfigureNotify )
				{
					//if ( e.xconfigure.width != w || e.xconfigure.height != h )
					{
						// OnSceneResize
						//w = e.xconfigure.width;
						//h = e.xconfigure.height;
						//std::cout << w << "x" << h << "\n";
						//m_gl_context->UpdateViewport(Size(w, h));
					}
				}
				else if ( e.type == MotionNotify )
				{
					// OnMouseMove
					//mx = e.xmotion.x;
					//my = e.xmotion.y;
				}
				else if ( e.type == ButtonPress || e.type == ButtonRelease )
				{
					//if ( e.type == ButtonPress )
					//{
					//	Lua::Execute(lua, "HandleMouseDown", mx, my, (int)e.xbutton.button);
					//}
					//else
					//{
					//	Lua::Execute(lua, "HandleMouseUp", mx, my, (int)e.xbutton.button);
					//}

					if ( e.type == ButtonRelease )
					{
						if ( e.xbutton.button == 1 )
						{
							//uint64_t now = TimeGetTime();
							//if ( lmbclick.time > 0 && lmbclick.time > now-350 )
							//{
							//	std::cout << "DblClick\n";
							//	lmbclick.time = 0;
							//	lmbclick.x = mx;
							//	lmbclick.y = my;
							//}
							//else
							//{
							//	std::cout << "Click\n";
							//	lmbclick.time = now;
							//	lmbclick.x = mx;
							//	lmbclick.y = my;
							//}
						}
					}
				}
				else if ( e.type == ClientMessage && e.xclient.data.l[0] == (long)wmDeleteMessage )
				{
					printf("Client message\n");
					m_proceed = false;
				}
				else if ( e.type == ClientMessage && e.xclient.format == 32 && e.xclient.data.l[0] == (long)this )
				{
					InvokationSentry* sentry = (InvokationSentry*)e.xclient.data.l[4];
					sentry->func();
					sentry->semaphore.Notify();
				}
			}
			else
			{
				usleep(1000);
			}
		}

		LOG(Trace, "Leaving X11-specific window thread function");
	}

	bool X11Window::Construct()
	{
		std::lock_guard<std::mutex> guard(m_lock);

		if ( !CreateWindowObject() || !CreateOpenGLContext() )
		{
			DestroyUnlocked();
			return false;
		}

		m_proceed = true;
		return true;
	}

	void X11Window::DestroyUnlocked()
	{
		DestroyOpenGLContext();
		DestroyWindowObject();
	}

	void X11Window::Destroy()
	{
		std::lock_guard<std::mutex> guard(m_lock);
		DestroyUnlocked();

		// Unblock possibly waiting client thread
		m_redraw_barrier.Notify();
	}

	bool X11Window::CreateWindowObject()
	{
		XInitThreads();
		setlocale(LC_ALL, "");

		m_private->display = XOpenDisplay(NULL);
		if ( m_private->display == NULL )
		{
			// TODO: LOG
			return false;
		}

		m_private->screen = DefaultScreen(m_private->display);

		m_private->visual = glXChooseVisual(m_private->display, m_private->screen, glx_double_buffered_attrs);

		if ( m_private->visual == NULL )
		{
			// TODO: LOG
			// Try fall back to single buffered (WTF?)
			m_private->visual = glXChooseVisual(m_private->display, m_private->screen, glx_single_buffered_attrs);
		}

		if ( m_private->visual == NULL )
		{
			// TODO: LOG
			return false;
		}

		// Log available OpenGL version
		int major, minor;
		glXQueryVersion(m_private->display, &major, &minor);
		LOG(Info, "Available OpenGL version: " << major << "." << minor);

		Colormap colormap = XCreateColormap
		(
			m_private->display,
			RootWindow(m_private->display, m_private->visual->screen),
			m_private->visual->visual,
			AllocNone
		);

		XSetWindowAttributes attrs;
		attrs.colormap = colormap;
		attrs.border_pixel = 0;
		attrs.event_mask = ExposureMask | KeyPressMask | ButtonPressMask | StructureNotifyMask;

		m_private->window = XCreateWindow
		(
			m_private->display,
			RootWindow(m_private->display, m_private->screen),
			0, 0,
			640, 480,
			0,
			m_private->visual->depth,
			InputOutput,
			m_private->visual->visual,
			CWBorderPixel | CWColormap | CWEventMask,
			&attrs
		);

		// Continue with GL
		m_private->glx = glXCreateContext(m_private->display, m_private->visual, 0, GL_TRUE);
		glXMakeCurrent(m_private->display, m_private->window, m_private->glx);

		// Continue with input
		if ( (m_private->im = XOpenIM(m_private->display, NULL, NULL, NULL)) == NULL )
		{
			LOG(Fatal, "Failed to open IM");
			return false;
		}

		XIMStyles *im_supported_styles;
		XIMStyle app_supported_styles;
		XIMStyle style;
		XIMStyle best_style;

		// Set flags for the styles an application can support
		app_supported_styles =
			XIMPreeditNone | XIMPreeditNothing | XIMPreeditArea |
			XIMStatusNone | XIMStatusNothing | XIMStatusArea;

		// Figure out which styles the IM can support
		XGetIMValues(m_private->im, XNQueryInputStyle, &im_supported_styles, NULL);

		// Look at each of the IM supported styles, and choise the "best" one that we can support
		best_style = 0;
		for( int i=0; i < im_supported_styles->count_styles; i++ )
		{
			style = im_supported_styles->supported_styles[i];
			if ( (style & app_supported_styles) == style ) // if we can handle it
			{
				best_style = ChooseBetterStyle(style, best_style);
			}
		}
		if (best_style == 0)
		{
			LOG(Fatal, "Can't find supported IM interaction style");
			return false;
		}

		XFree(im_supported_styles);

		// Input context
		m_private->ic = XCreateIC(m_private->im, XNInputStyle, best_style, XNClientWindow, m_private->window, NULL);
		if ( m_private->ic == NULL )
		{
			LOG(Fatal, "Failed to create input context");
			return false;
		}

		long im_event_mask;
		XGetICValues(m_private->ic, XNFilterEvents, &im_event_mask, NULL);
		XSetICFocus(m_private->ic);
		XSelectInput(m_private->display, m_private->window, ExposureMask | KeyPressMask | StructureNotifyMask | im_event_mask);

		return true;
	}

	bool X11Window::CreateOpenGLContext()
	{
		// ...
		return true;
	}

	void X11Window::DestroyWindowObject()
	{
		// TODO: Release IC

		// TODO: Release IM

		// TODO: Unbind glX
		glXMakeCurrent(m_private->display, None, NULL);

		// TODO: Destroy glX context
		glXDestroyContext(m_private->display, m_private->glx);

		// TODO: Destroy window
		XDestroyWindow(m_private->display, m_private->window);

		if ( m_private->display != NULL )
		{
			//XCloseDisplay(m_private->display);
			m_private->display = NULL;
		}
	}

	void X11Window::DestroyOpenGLContext()
	{
		// ...
	}

	void X11Window::ReportInput(const Keystroke& keystroke)
	{
		InvokeCallback(m_on_input, keystroke);
	}
}

#endif

